__all__ = ()

from datetime import datetime

from scarletio import Compound

from ...channel import Channel
from ...core import SCHEDULED_EVENTS
from ...http import DiscordHTTPClient
from ...scheduled_event import PrivacyLevel, ScheduledEvent, ScheduledEventEntityType, ScheduledEventStatus
from ...user import ClientUserBase, User
from ...utils import datetime_to_timestamp, log_time_converter
from ..request_helpers import (
    get_channel_id, get_guild_id, get_scheduled_event_guild_id_and_id, get_scheduled_event_and_guild_id_and_id
)


class ClientCompoundScheduledEventEndpoints(Compound):
    
    http : DiscordHTTPClient
    
    
    async def scheduled_event_create(
        self, guild, name, start, *, description=None, end=None, privacy_level=PrivacyLevel.guild_only,
        location=None, stage=None, voice=None, reason=None
    ):
        """
        Creates a guild scheduled events.
        
        To define, where the event will take place, use the `location`, `stage`, `voice` parameters.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild, where to create the scheduled event.
        name : `str`
            The event's name. It's length can be in range [1:100].
        start : `datetime``
            When the event will start.
        description : `None`, `str` = `None`, Optional (Keyword only)
            The event's description. It's length can be in range [0:1000].
        end : `None`, `datetime` = `None`, Optional (Keyword only)
            When the event will end.
        privacy_level : ``PrivacyLevel``, `int` = `None`, Optional (Keyword only)
            The privacy level of the event. Whether it is global or guild only.
        location : `None`, `str` = `None`, Optional (Keyword only)
            The location, where the event will take place.
        stage : `None`, ``Channel``, `int` = `None`, Optional (Keyword only)
            The stage channel, where the event will take place.
        voice : `None`, ``Channel``, `int` = `None`, Optional (Keyword only)
            The voice channel, where the event will take place.
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Shows up at the respective guild's audit logs.
        
        Returns
        -------
        scheduled_event : ``ScheduledEvent``
            The created event.
        
        Raises
        ------
        TypeError
            - If `guild` is neither ``Guild``, nor `int`.
            - If `privacy_level` is neither ``PrivacyLevel``, nor `int`.
            - If `stage` is neither ``Channel``, nor `int`.
            - If `voice` is neither ``Channel``, nor `int`.
            - If neither `location`, `stage`, `voice` parameters is given.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            - If `name` is not `str`.
            - If `name`'s length is out of the expected range.
            - If `description is neither `None` nor `str`.
            - If `description`'s length is out of the expected range.
            - If `location` is not `str`.
            - If `start` is not `datetime`.
            - If `end `is neither `None`, nor `datetime`.
        """
        guild_id = get_guild_id(guild)
        
        if __debug__:
            if not isinstance(name, str):
                raise AssertionError(
                    f'`name` can be `str`, got {name.__class__.__name__}; {name!r}.'
                )
            
            name_length = len(name)
            if (name_length < 1) or (name_length > 100):
                raise AssertionError(
                    f'`name` length can be in range [1:100], got {name_length!r}; {name!r}.'
                )
        
        if __debug__:
            if not isinstance(start, datetime):
                raise AssertionError(
                    f'`start` can be `datetime`, got {start.__class__.__name__}; {start!r}.'
                )
            
            if (end is not None) and (not isinstance(end, datetime)):
                raise AssertionError(
                    f'`end` can be `None`, `datetime`, got {end.__class__.__name__}; {end!r}.'
                )
        
        if isinstance(privacy_level, PrivacyLevel):
            privacy_level_value = privacy_level.value
        elif isinstance(privacy_level, int):
            privacy_level_value = privacy_level
        else:
            raise TypeError(
                f'`privacy_level` can be `{PrivacyLevel.__name__}` `int`, got '
                f'{privacy_level.__class__.__name__}; {privacy_level!r}.'
            )
        
        if (description is not None):
            if __debug__:
                if (not isinstance(description, str)):
                    raise AssertionError(
                        f'`description` can be `None`, `str`, got {description.__class__.__name__}, {description!r}.'
                    )
                
                description_length = len(description)
                if description_length > 1000:
                    raise AssertionError(
                        f'description length can be in range [0:1000], got {description_length!r}; {description!r}.'
                    )
            
            if not description:
                description = None
        
        if (location is not None):
            if __debug__:
                if not isinstance(location, str):
                    raise AssertionError(
                        f'`location` can be `str`, got {location.__class__.__name__}; {location!r}.'
                    )
            
            channel_id = None
            entity_metadata = {'location': location}
            entity_type = ScheduledEventEntityType.location
        
        elif (stage is not None):
            channel_id = get_channel_id(stage, Channel.is_guild_stage)
            entity_metadata = None
            entity_type = ScheduledEventEntityType.stage
        
        elif (voice is not None):
            channel_id = get_channel_id(voice, Channel.is_guild_voice)
            entity_metadata = None
            entity_type = ScheduledEventEntityType.voice
            
        else:
            raise TypeError(
                f'Either `location`, `stage`, `voice` parameters are required.'
            )
        
        data = {
            'name' : name,
            'privacy_level': privacy_level_value,
            'channel_id': channel_id,
            'entity_metadata': entity_metadata,
            'entity_type': entity_type.value,
            'scheduled_start_time': datetime_to_timestamp(start),
        }
        
        if (description is not None):
            data['description'] = description
        
        if (end is not None):
            data['scheduled_end_time'] = datetime_to_timestamp(end)
        
        data = await self.http.scheduled_event_create(guild_id, data, reason)
        return ScheduledEvent(data)
    
    # In theory you can edit the target entity is as well, but we will ignore it for now.
    
    async def scheduled_event_edit(
        self, scheduled_event, *, name=..., description=..., start=..., end=..., privacy_level=..., status=...,
        location=..., stage=..., voice=..., reason=None
    ):
        """
        Edits the given scheduled event.
        
        This method is a coroutine.
        
        Parameters
        ----------
        scheduled_event : ``ScheduledEvent``, `tuple`, (`int`, `int`)
            The scheduled event to edit.
        name : `str`, Optional (Keyword only)
            The new name of the scheduled event. It's length can be in range [1:100].
        description : `None`, `str`, Optional (Keyword only)
            The new description of the scheduled event. It's length can be in range [0:1000].
            
            Pass it as `None` to remove the old one.
        
        start : `datetime`, Optional (Keyword only)
            The new start of the scheduled event.
        
        end : `None`, `datetime`
            The end of the of the scheduled event.
            
            Pass it as `None` to remove the old end.
        
        privacy_level : ``PrivacyLevel``, `int`, Optional (Keyword only)
            The privacy level of the event. Whether it is global or guild only.
        
        status : `str`, ``ScheduledEventStatus``, Optional (Keyword only)
            Thew new status of the scheduled event.
        
        location : `str`, Optional (Keyword only)
            The new location, where the event will take place.
        
        stage : ``Channel``, `int`, Optional (Keyword only)
            The new stage channel, where the event will take place.
        
        voice : ``Channel``, `int`, Optional (Keyword only)
            The new voice channel, where the event will take place.
        
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Shows up at the respective guild's audit logs.
        
        Raises
        ------
        TypeError
            - If `scheduled_event` is neither ``ScheduledEvent``, nor `tuple` (`int`, `int`).
            - If `privacy_level` is neither ``PrivacyLevel``, nor `int`.
            - If `stage` is neither ``Channel``, nor `int`.
            - If `voice` is neither ``Channel``, nor `int`.
            - If `privacy_level` is neither ``PrivacyLevel``, nor `int`.
            - If `status` is neither ``ScheduledEventStatus``, nor `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            - If `name` is not `str`.
            - If `name`'s length is out of the expected range.
            - If `description is neither `None` nor `str`.
            - If `description`'s length is out of the expected range.
            - If `location` is not `str`.
            - If `start` is not `datetime`.
            - If `end `is neither `None`, nor `datetime`.
        """
        guild_id, scheduled_event_id = get_scheduled_event_guild_id_and_id(scheduled_event)
        
        data = {}
        
        if (name is not ...):
            if __debug__:
                if not isinstance(name, str):
                    raise AssertionError(
                        f'`name` can be `str`, got {name.__class__.__name__}; {name!r}.'
                    )
                
                name_length = len(name)
                if (name_length < 1) or (name_length > 100):
                    raise AssertionError(
                        f'`name` length can be in range [1:100], got {name_length!r}; {name!r}.'
                    )
            
            data['name'] = name
        
        
        if (description is not ...):
            if description is None:
                description = ''
            else:
                if __debug__:
                    if (not isinstance(description, str)):
                        raise AssertionError(
                            f'`description` can be `None`, `str`, got '
                            f'{description.__class__.__name__}; {description!r}.'
                        )
                    
                    description_length = len(description)
                    if description_length > 1000:
                        raise AssertionError(
                            f'description length can be in range [0:1000], got '
                            f'{description_length!r}; {description!r}.'
                        )
            
            data['description'] = description
        
        if (start is not ...):
            if __debug__:
                if not isinstance(start, datetime):
                    raise AssertionError(
                        f'`start` can be `datetime`, got {start.__class__.__name__}; {start!r}.'
                    )
            
            data['scheduled_start_time'] = datetime_to_timestamp(start)
        
        if (end is not ...):
            if __debug__:
                if (end is not None) and (not isinstance(end, datetime)):
                    raise AssertionError(
                        f'`end` can be `None`, `datetime`, got {end.__class__.__name__}; {end!r}.'
                    )
            
            data['scheduled_end_time'] = None if end is None else datetime_to_timestamp(end)
        
        if (privacy_level is not ...):
            if isinstance(privacy_level, PrivacyLevel):
                privacy_level_value = privacy_level.value
            elif isinstance(privacy_level, int):
                privacy_level_value = privacy_level
            else:
                raise TypeError(
                    f'`privacy_level` can be `{PrivacyLevel.__name__}`, `int`, got '
                    f'{privacy_level.__class__.__name__}; {privacy_level!r}.'
                )
            
            data['privacy_level'] = privacy_level_value
        
        if (status is not ...):
            if isinstance(status, ScheduledEventStatus):
                status_value = status.value
            elif isinstance(status, int):
                status_value = status
            else:
                raise TypeError(
                    f'`status` can be `{ScheduledEventStatus.__name__}`, `int` , got '
                    f'{status.__class__.__name__}; {status!r}.'
                )
            
            data['status'] = status_value
        
        if (location is not ...) or (stage is not ...) or (voice is not ...):
            if (location is not ...):
                if __debug__:
                    if not isinstance(location, str):
                        raise AssertionError(
                            f'`location` can be `str`, got {location.__class__.__name__}; {location!r}.'
                        )
                
                channel_id = None
                entity_metadata = {'location': location}
                entity_type = ScheduledEventEntityType.location
            
            elif (stage is not ...):
                channel_id = get_channel_id(stage, Channel.is_guild_stage)
                entity_metadata = None
                entity_type = ScheduledEventEntityType.stage
            
            # elif (voice is not ...):
            else:
                channel_id = get_channel_id(voice, Channel.is_guild_voice)
                entity_metadata = None
                entity_type = ScheduledEventEntityType.voice
            
            data['channel_id'] = channel_id
            data['entity_metadata'] = entity_metadata
            data['entity_type'] = entity_type.value
        
        
        if data:
            await self.http.scheduled_event_edit(guild_id, scheduled_event_id, data, reason)
    
    
    async def scheduled_event_delete(self, scheduled_event):
        """
        Edits the given scheduled event.
        
        This method is a coroutine.
        
        Parameters
        ----------
        scheduled_event : ``ScheduledEvent``, `tuple` (`int`, `int`)
            The scheduled event to edit.
        
        Raises
        ------
        TypeError
            If `scheduled_event` is neither ``ScheduledEvent``, nor `tuple` (`int`, `int`).
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id, scheduled_event_id = get_scheduled_event_guild_id_and_id(scheduled_event)
        await self.http.scheduled_event_delete(guild_id, scheduled_event_id)
    
    
    async def scheduled_event_get(self, scheduled_event, *, force_update=False):
        """
        Gets the given scheduled event.
        
        This method is a coroutine.
        
        Parameters
        ----------
        scheduled_event : ``ScheduledEvent``, `tuple` `int` and `int`
            The scheduled event to get.
        force_update : `bool` = `False`, Optional (Keyword only)
            Whether the scheduled event should be requested even if it supposed to be up to date.
        
        Returns
        -------
        scheduled_event : ``ScheduledEvent``
        
        Raises
        ------
        TypeError
            If `scheduled_event` is neither ``ScheduledEvent``, nor `tuple` (`int`, `int`) instance.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        scheduled_event, guild_id, scheduled_event_id = get_scheduled_event_and_guild_id_and_id(scheduled_event)
        if (scheduled_event is None) or force_update:
            data = await self.http.scheduled_event_get(guild_id, scheduled_event_id, {'with_user_count', None})
        
            if scheduled_event is None:
                try:
                    scheduled_event = SCHEDULED_EVENTS[scheduled_event_id]
                except KeyError:
                    scheduled_event = ScheduledEvent(data)
                else:
                    scheduled_event._update_attributes(data)
                    scheduled_event._update_counts_only(data)
            else:
                scheduled_event = ScheduledEvent(data)
        
        return scheduled_event
    
    
    async def scheduled_event_get_all_guild(self, guild):
        """
        Gets the given guild's scheduled event.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild to get it's scheduled of.
        
        Returns
        -------
        scheduled_events : `list` of ``ScheduledEvent``
        
        Raises
        ------
        TypeError
            If `guild` is neither ``Guild`` nor `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id = get_guild_id(guild)
        scheduled_event_datas = await self.http.scheduled_event_get_all_guild(guild_id, {'with_user_count': True})
        return [ScheduledEvent(scheduled_event_data) for scheduled_event_data in scheduled_event_datas]
    
    
    async def scheduled_event_user_get_chunk(self, scheduled_event, *, after=None, before=None, limit=None):
        """
        Requests a chunk user subscribed to a scheduled event.
        
        This method is a coroutine.
        
        Parameters
        ----------
        scheduled_event : ``ScheduledEvent``, `tuple` `int` and `int`
            The scheduled event to get.
        after : `None`, `int`, ``DiscordEntity``, `datetime` = `None`, Optional (Keyword only)
            The timestamp after the subscribed users were created.
        before : `None`, `int`, ``DiscordEntity``, `datetime` = `None`, Optional (Keyword only)
            The timestamp before the subscribed users were created.
        limit : `None`, `int` = `None`, Optional (Keyword only)
            The amount of scheduled event users to request. Can be between 1 and 100.
        
        Returns
        -------
        users : `list` of ``ClientUserBase``
        
        Raises
        ------
        TypeError
            - If `scheduled_event` is neither ``ScheduledEvent``, nor `tuple` (`int`, `int`) instance.
            - If `after`, `before` was passed with an unexpected type.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            - If `limit` was not given as `int`.
            - If `limit` is out of range [1:100].
        """
        guild_id, scheduled_event_id = get_scheduled_event_guild_id_and_id(scheduled_event)
        
        if limit is None:
            limit = 100
        else:
            if __debug__:
                if not isinstance(limit, int):
                    raise AssertionError(
                        f'`limit` can be `int`, got {limit.__class__.__name__}; {limit!r}.'
                    )
                
                if (limit < 1) or (limit > 100):
                    raise AssertionError(
                        f'`limit` is out from the expected [1:100] range, got {limit!r}.'
                    )
        
        query_parameters = {}
        
        if limit != 100:
            query_parameters['limit'] = limit
        
        if (after is not None):
            query_parameters['after'] = log_time_converter(after)
        
        if (before is not None):
            query_parameters['before'] = log_time_converter(before)
        
        if guild_id:
            query_parameters['with_member'] = True
        
        scheduled_event_user_datas = await self.http.scheduled_event_user_get_chunk(
            guild_id, scheduled_event_id, query_parameters
        )
        
        users = []
        for scheduled_event_user_data in scheduled_event_user_datas:
            user_data = scheduled_event_user_data['user']
            user = User.from_data(user_data, user_data.get('member', None), guild_id)
            users.append(user)
        
        return users
    
    
    async def scheduled_event_user_get_all(self, scheduled_event):
        """
        Requests a chunk user subscribed to a scheduled event.
        
        This method is a coroutine.
        
        Parameters
        ----------
        scheduled_event : ``ScheduledEvent``, `tuple` `int` and `int`
            The scheduled event to get.
        
        Returns
        -------
        users : `list` of ``ClientUserBase``
        
        Raises
        ------
        TypeError
            - If `scheduled_event` is neither ``ScheduledEvent``, nor `tuple` (`int`, `int`) instance.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id, scheduled_event_id = get_scheduled_event_guild_id_and_id(scheduled_event)
        
        query_parameters = {'after': 0}
        
        if guild_id:
            query_parameters['with_member'] = True
        
        users = []
        
        while True:
            scheduled_event_user_datas = await self.http.scheduled_event_user_get_chunk(
                guild_id,
                scheduled_event_id,
                query_parameters,
            )
            
            for scheduled_event_user_data in scheduled_event_user_datas:
                user_data = scheduled_event_user_data['user']
                user = User.from_data(user_data, user_data.get('member', None), guild_id)
                users.append(user)
            
            if len(scheduled_event_user_datas) < 100:
                break
            
            query_parameters['after'] = users[-1].id
        
        return users



from .application_command_autocomplete_interaction import *
from .application_command_interaction import *
from .component_interaction import *
from .form_submit_interaction import *
from .interaction_event import *
from .interaction_field_base import *
from .preinstanced import *

__all__ = (
    *application_command_autocomplete_interaction.__all__,
    *application_command_interaction.__all__,
    *component_interaction.__all__,
    *form_submit_interaction.__all__,
    *interaction_event.__all__,
    *interaction_field_base.__all__,
    *preinstanced.__all__,
)



__all__ = ('set_debug_logger', )

import sys

from .. import __package__ as PACKAGE_NAME


UNIQUE_MESSAGES = set()


def call_debug_logger(message, unique):
    """
    Calls debug logger if applicable.
    
    Parameters
    ----------
    message : `str`
        The message to show.
    unique : `bool`
        Whether the message should not be shown if already was.
    """
    if unique:
        if message in UNIQUE_MESSAGES:
            return
        
        UNIQUE_MESSAGES.add(message)
    
    _debug_logger(message)


def set_debug_logger(debug_logger):
    """
    Sets the given function as debug handler.
    
    Parameters
    ----------
    debug_logger : `callable`
        Any callable accepting 1 string parameter.
    """
    global _debug_logger
    _debug_logger = debug_logger


def default_debug_logger(message):
    """
    Default debug logger.
    
    Puts `DEBUG` prefix before each logged line and writes it to stdout.
    
    To set a new debug logger use ``set_debug_logger``.
    
    Parameters
    ----------
    message : `str`
        The message to show.
    """
    lines = message.splitlines()
    output_parts = []
    for line in lines:
        output_parts.append(PACKAGE_NAME)
        output_parts.append(' DEBUG: ')
        output_parts.append(line)
        output_parts.append('\n')
    
    output = ''.join(output_parts)
    
    sys.stdout.write(output)


_debug_logger = default_debug_logger



__all__ = ()


import warnings
from collections import deque
from time import time as time_now

from scarletio import Compound, Task, WaitTillAll, WaitTillFirst

from ...allowed_mentions import parse_allowed_mentions
from ...bases import maybe_snowflake, maybe_snowflake_pair
from ...channel import Channel, MessageIterator, message_relative_index
from ...core import CHANNELS, KOKORO, MESSAGES
from ...exceptions import DiscordException, ERROR_CODES
from ...http import DiscordHTTPClient
from ...message import Message, MessageFlag, MessageReference, MessageRepr
from ...message.utils import process_message_chunk
from ...permission.permission import PERMISSION_MASK_MANAGE_MESSAGES, PERMISSION_MASK_READ_MESSAGE_HISTORY
from ...sticker import Sticker
from ...utils import DISCORD_EPOCH, log_time_converter

from ..functionality_helpers import (
    MultiClientMessageDeleteSequenceSharder, _message_delete_multiple_private_task, _message_delete_multiple_task
)

from ..request_helpers import (
    add_file_to_message_data, get_channel_and_id, get_channel_id, get_channel_id_and_message_id, get_components_data,
    get_message_and_channel_id_and_message_id, validate_content_and_embed, validate_message_to_delete
)


MESSAGE_FLAG_VALUE_SUPPRESS_EMBEDS = MessageFlag().update_by_keys(embeds_suppressed=True)


class ClientCompoundMessageEndpoints(Compound):
    
    http : DiscordHTTPClient
    id : int
    
    
    async def message_get_chunk(self, channel, limit=100, *, after=None, around=None, before=None):
        """
        Requests messages from the given text channel. The `after`, `around` and the `before` parameters are mutually
        exclusive and they can be `int`, or as a ``DiscordEntity`` or as a `datetime` object.
        If there is at least 1 message overlap between the received and the loaded messages, the wrapper will chain
        the channel's message history up. If this happens the channel will get on a queue to have it's messages again
        limited to the default one, but requesting old messages more times, will cause it to extend.
        
        This method is a coroutine.
        
        Parameters
        ----------
        channel : ``Channel``, `int`
            The channel from where we want to request the messages.
        limit : `int` = `100`, Optional
            The amount of messages to request. Can be between 1 and 100.
        after : `None`, `int`, ``DiscordEntity``, `datetime` = `None`, Optional (Keyword only)
            The timestamp after the requested messages were created.
        around : `None`, `int`, ``DiscordEntity``, `datetime` = `None`, Optional (Keyword only)
            The timestamp around the requested messages were created.
        before : `None`, `int`, ``DiscordEntity``, `datetime` = `None`, Optional (Keyword only)
            The timestamp before the requested messages were created.
        
        Returns
        -------
        messages : `list` of ``Message``
        
        Raises
        ------
        TypeError
            - If `channel` was not given neither as ``Channel`` nor `int`.
            - If `after`, `around`, `before` was passed with an unexpected type.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            - If `limit` was not given as `int`.
            - If `limit` is out of range [1:100].
        
        See Also
        --------
        - ``.message_get_chunk_from_zero`` : Familiar to this method, but it requests only the newest messages of the
            channel and makes sure they are chained up with the channel's message history.
        - ``.message_get_at_index`` : A top-level method to get a message at the specified index at the given channel.
            Usually used to load the channel's message history to that point.
        - ``.message_get_all_in_range`` : A top-level method to get all the messages till the specified index at the
            given channel.
        - ``.message_iterator`` : An iterator over a channel's message history.
        """
        channel, channel_id = get_channel_and_id(channel, Channel.is_in_group_messageable)
        
        if __debug__:
            if not isinstance(limit, int):
                raise AssertionError(
                    f'`limit` can be `int`, got {limit.__class__.__name__}; {limit!r}.'
                )
            
            if (limit < 1) or (limit > 100):
                raise AssertionError(
                    f'`limit` is out from the expected [1:100] range, got {limit!r}.'
                )
        
        data = {'limit': limit}
        
        if (after is not None):
            data['after'] = log_time_converter(after)
        
        if (around is not None):
            data['around'] = log_time_converter(around)
        
        if (before is not None):
            data['before'] = log_time_converter(before)
        
        # Set some collection delay.
        if (channel is not None):
            channel._add_message_collection_delay(60.0)
        
        message_datas = await self.http.message_get_chunk(channel_id, data)
        return process_message_chunk(message_datas, channel)
    
    
    # If you have 0-1 messages at a channel, and you wanna store the messages. The other wont store it, because it
    # wont see anything what allows channeling.
    async def message_get_chunk_from_zero(self, channel, limit=100):
        """
        If the `channel` has `1` or less messages loaded use this method instead of ``.message_get_chunk`` to request
        the newest messages there, because this method makes sure, the returned messages will be chained at the
        channel's message history.
        
        This method is a coroutine.
        
        Parameters
        ----------
        channel : ``Channel``
            The channel from where we want to request the messages.
        limit : `int` = `100`, Optional
            The amount of messages to request. Can be between 1 and 100.
        
        Returns
        -------
        messages : `list` of ``Message`` objects
        
        Raises
        ------
        TypeError
            If `channel` was not given neither as ``Channel`` nor `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            - If `limit` was not given as `int`.
            - If `limit` is out of range [1:100].
        """
        channel, channel_id = get_channel_and_id(channel, Channel.is_in_group_messageable)
        
        if __debug__:
            if not isinstance(limit, int):
                raise AssertionError(
                    f'`limit` can be `int`, got {limit.__class__.__name__}; {limit!r}.'
                )
            
            if (limit < 1) or (limit > 100):
                raise AssertionError(
                    f'`limit` is out from the expected [1:100] range, got {limit!r}.'
                )
        
        # Set some collection delay.
        if (channel is not None):
            channel._add_message_collection_delay(60.0)
        
        data = {'limit': limit, 'before': 9223372036854775807}
        data = await self.http.message_get_chunk(channel_id, data)
        if data:
            if (channel is not None):
                # Call this method first, so the channel's messages will be set even if message caching is at 0
                channel._maybe_increase_queue_size()
                
                channel._create_new_message(data[0])
            
            messages = process_message_chunk(data, channel)
        else:
            messages = []
        
        return messages
    
    
    async def message_get(self, message, *positional_parameters, force_update=False):
        """
        Requests a specific message by it's id at the given `channel`.
        
        This method is a coroutine.
        
        Parameters
        ----------
        message : ``Message``, ``MessageRepr``, ``MessageReference``, `tuple` (`int`, `int`)
            The message to get, or a `channel-id`, `message-id` tuple representing it.
        force_update : `bool` = `False`, Optional (Keyword only)
            Whether the scheduled event should be requested even if it supposed to be up to date.
        
        Returns
        -------
        message : ``Message``
        
        Raises
        ------
        TypeError
            - If `message`'s type is neither ``Message``, nor `tuple` (`int`, `int`).
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        if positional_parameters:
            warnings.warn(
                f'`{self.__class__.__name__}.message_get` parameters are modified from `channel + message_id` to '
                f'`Message` / `(channel_id, message_id`). The old usage is deprecated and will be removed in 2022 '
                f'December.'
            )
            
            channel, message_id = message, *positional_parameters
                
            channel_id = get_channel_id(channel, Channel.is_in_group_messageable)
            
            message_id_value = maybe_snowflake(message_id)
            if message_id_value is None:
                raise TypeError(
                    f'`message_id` can be `int`, got {message_id.__class__.__name__}; {message_id!r}.'
                )
            
            message_id = message_id_value
        
        else:
            if isinstance(message, Message):
                message_id = message.id
                channel_id = message.channel_id
            
            elif isinstance(message, MessageRepr):
                message_id = message.id
                channel_id = message.channel_id
                message = None
                
            elif isinstance(message, MessageReference):
                message_id = message.message_id
                channel_id = message.channel_id
                message = None
            
            else:
                snowflake_pair = maybe_snowflake_pair(message)
                if (snowflake_pair is not None):
                    channel_id, message_id = snowflake_pair
                    message = None
                
                else:
                    raise TypeError(
                        f'`message` can be `{Message.__name__}`, `{MessageRepr.__name__}`, '
                        f'`{MessageReference.__name__}`, `tuple` (`int`, `int`), '
                        f'got {message.__class__.__name__}; {message!r}.'
                    )
        
        message_data = await self.http.message_get(channel_id, message_id)
        
        if force_update:
            if message is None:
                message = MESSAGES.get(message_id, None)
            
            if (message is None):
                message = Message(message_data)
            
            else:
                message._set_attributes(message_data)
        
        else:
            message = Message(message_data)
        
        return message
    
    
    async def message_create(
        self, channel, content=None, *, allowed_mentions=...,  components=None, embed=None, file=None, nonce=None,
        reply_fail_fallback=False,  sticker=None, suppress_embeds=False, tts=False
    ):
        """
        Creates and returns a message at the given `channel`. If there is nothing to send, then returns `None`.
        
        This method is a coroutine.
        
        Parameters
        ----------
        channel : ``Channel``, `int`, ``Message``, ``MessageRepr``, ``MessageReference``,
                `tuple` (`int`, `int`)
            The text channel where the message will be sent, or the message on what you want to reply.
        
        content : `None`, `str`, ``EmbedBase``, `Any` = `None`, Optional
            The message's content if given. If given as `str` or empty string, then no content will be sent, meanwhile
            if any other non `str`, ``EmbedBase`` is given, then will be casted to string.
            
            If given as ``EmbedBase``, then is sent as the message's embed.
        
        allowed_mentions : `None`,  `str`, ``UserBase``, ``Role``, `list` of (`str`, ``UserBase``, ``Role`` )
                , Optional (Keyword only)
            Which user or role can the message ping (or everyone). Check ``parse_allowed_mentions`` for details.
        
        components : `None`, ``ComponentBase``, (`tuple`, `list`) of (``ComponentBase``, (`tuple`, `list`) of
                ``ComponentBase``) = `None`, Optional (Keyword only)
            Components attached to the message.
            
            > `components` do not count towards having any content in the message.
        
        embed : ``EmbedBase``, `list` of ``EmbedBase`` = `None`, Optional (Keyword only)
            The embedded content of the message.
            
            If `embed` and `content` parameters are both given as  ``EmbedBase``, then `TypeError` is raised.
        
        file : `None`, `Any` = `None`, Optional (Keyword only)
            A file or files to send. Check ``create_file_form`` for details.
        
        reply_fail_fallback : `bool` = `False`, Optional (Keyword only)
            Whether normal message should be sent if the referenced message is deleted. Defaults to `False`.
        
        nonce : `None`, `str` = `None`, Optional (Keyword only)
            Used for optimistic message sending. Will shop up at the message's data.
        
        sticker : `None`, ``Sticker``, `int`, (`list`, `set`, `tuple`) of (``Sticker``, `int`) = `None` \
                , Optional (Keyword only)
            Sticker or stickers to send within the message.
        
        suppress_embeds : `bool` = `False`, Optional (Keyword only)
            Whether the message's embeds should be suppressed initially.
        
        tts : `bool` = `False`, Optional (Keyword only)
            Whether the message is text-to-speech.
        
        Returns
        -------
        message : `None`, ``Message``
            Returns `None` if there is nothing to send.
        
        Raises
        ------
        TypeError
            - If `embed` was not given neither as ``EmbedBase`` nor as `list`, `tuple` of ``EmbedBase``-s.
            - If `allowed_mentions` contains an element of invalid type.
            - `content` parameter was given as ``EmbedBase``, meanwhile `embed` parameter was given as well.
            - If invalid file type would be sent.
            - If `channel`'s type is incorrect.
            - If `sticker` was not given neither as `None`, ``Sticker``, `int`, (`list`, `tuple`) of \
                (``Sticker``, `int).
            - If `components` type is incorrect.
        ValueError
            - If `allowed_mentions`'s elements' type is correct, but one of their value is invalid.
            - If more than `10` files would be sent.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            - If `tts` was not given as `bool`.
            - If `nonce` was not given neither as `None` nor as `str`.
            - If `reply_fail_fallback` was not given as `bool`.
            - If `embed` contains a non ``EmbedBase`` element.
            - If `suppress_embeds` is not `bool`.
            - If both `content` and `embed` fields are embeds.
        
        See Also
        --------
        ``.webhook_message_create`` : Sending a message with a ``Webhook``.
        """
        
        # Channel check order:
        # 1.: Channel -> channel
        # 2.: Message -> channel + reply
        # 3.: int (str) -> channel
        # 4.: MessageRepr -> channel + reply
        # 5.: MessageReference -> channel + reply
        # 6.: `tuple` (`int`, `int`) -> channel + reply
        # 7.: raise
        
        while True:
            if isinstance(channel, Channel):
                if channel.is_in_group_messageable() or channel.partial:
                    message_id = None
                    channel_id = channel.id
                    break
            
            elif isinstance(channel, Message):
                message_id = channel.id
                channel_id = channel.channel_id
                channel = CHANNELS.get(channel_id, None)
                break
            
            else:
                channel_id = maybe_snowflake(channel)
                if (channel_id is not None):
                    message_id = None
                    channel = CHANNELS.get(channel_id, None)
                    break
                
                elif isinstance(channel, MessageRepr):
                    message_id = channel.id
                    channel_id = channel.channel_id
                    channel = CHANNELS.get(channel_id, None)
                    break
                
                elif isinstance(channel, MessageReference):
                    message_id = channel.message_id
                    channel_id = channel.channel_id
                    channel = CHANNELS.get(channel_id, None)
                    break
                
                else:
                    snowflake_pair = maybe_snowflake_pair(channel)
                    if snowflake_pair is not None:
                        channel_id, message_id = snowflake_pair
                        channel = CHANNELS.get(channel_id, None)
                        break
            
            raise TypeError(
                f'`channel` can be a messageable channel, `{Message.__name__}`, '
                f'`{MessageRepr.__name__}`, `{MessageReference.__name__}`, `int`, `tuple` (`int`, `int`), '
                f'got {channel.__class__.__name__}; {channel!r}.'
            )
            
        
        content, embed = validate_content_and_embed(content, embed, False)
        
        # Sticker check order:
        # 1.: None -> None
        # 2.: Sticker -> [sticker.id]
        # 3.: int (str) -> [sticker]
        # 4.: (list, tuple) of (Sticker, int (str)) -> [sticker.id / sticker, ...] / None
        # 5.: raise
        
        if sticker is None:
            sticker_ids = None
        else:
            sticker_ids = []
            if isinstance(sticker, Sticker):
                sticker_id = sticker.id
                sticker_ids.append(sticker_id)
            else:
                sticker_id = maybe_snowflake(sticker)
                if sticker_id is None:
                    if isinstance(sticker, (list, tuple)):
                        for sticker_element in sticker:
                            if isinstance(sticker_element, Sticker):
                                sticker_id = sticker_element.id
                            else:
                                sticker_id = maybe_snowflake(sticker_element)
                                if sticker_id is None:
                                    raise TypeError(
                                        f'`sticker` can contain only `{Sticker.__name__}`, `int` elements, '
                                        f'got {sticker_element.__class__.__name__}; {sticker_element!r}; '
                                        f'sticker={sticker!r}.'
                                    )
                            
                            sticker_ids.append(sticker_id)
                        
                        if not sticker_ids:
                            sticker_ids = None
                    else:
                        raise TypeError(
                            f'`sticker` can be `None`, `{Sticker.__name__}`, `int`, '
                            f'(`list`, `tuple`) of (`{Sticker.__name__}`, `int`), got '
                            f'{sticker.__class__.__name__}; {sticker!r}.'
                        )
                else:
                    sticker_ids.append(sticker_id)
        
        components = get_components_data(components, False)
        
        if __debug__:
            
            if (nonce is not None) and (not isinstance(nonce, str)):
                raise AssertionError(
                    f'`nonce` can be `None`, `str`, got {nonce.__class__.__name__}; {nonce!r}.'
                )
            
            if not isinstance(reply_fail_fallback, bool):
                raise AssertionError(
                    f'`reply_fail_fallback` can be `bool`, got {reply_fail_fallback.__class__.__name__}; '
                    f'{reply_fail_fallback!r}.')
            
            if not isinstance(suppress_embeds, bool):
                raise AssertionError(
                    f'`suppress_embeds` can be `bool`, got {suppress_embeds.__class__.__name__}; {suppress_embeds!r}.'
                )
        
            if not isinstance(tts, bool):
                raise AssertionError(
                    f'`tts` can be `bool`, got {tts.__class__.__name__}, {tts!r}.'
                )
        
        # Build payload
        message_data = {}
        contains_content = False
        
        if (content is not None):
            message_data['content'] = content
            contains_content = True
        
        if (embed is not None):
            message_data['embeds'] = [embed.to_data() for embed in embed]
            contains_content = True
        
        if (sticker_ids is not None):
            message_data['sticker_ids'] = sticker_ids
            contains_content = True
        
        if (components is not None):
            message_data['components'] = components
        
        if tts:
            message_data['tts'] = True
        
        if (nonce is not None):
            message_data['nonce'] = nonce
        
        if (allowed_mentions is not ...):
            message_data['allowed_mentions'] = parse_allowed_mentions(allowed_mentions)
        
        if (message_id is not None):
            message_reference_data = {'message_id': message_id}
            
            if reply_fail_fallback:
                message_reference_data['fail_if_not_exists'] = False
            
            message_data['message_reference'] = message_reference_data
        
        if suppress_embeds:
            message_data['flags'] = MESSAGE_FLAG_VALUE_SUPPRESS_EMBEDS
        
        message_data = add_file_to_message_data(message_data, file, contains_content, False)
        if message_data is None:
            return
        
        message_data = await self.http.message_create(channel_id, message_data)
        if (channel is not None):
            return channel._create_new_message(message_data)
    

    async def message_edit(
        self, message, content=..., *, embed=..., file=..., allowed_mentions=..., components=..., suppress=...,
        suppress_embeds=...
    ):
        """
        Edits the given `message`.
        
        This method is a coroutine.
        
        Parameters
        ----------
        message : ``Message``, ``MessageRepr``, ``MessageReference``, `tuple` (`int`, `int`)
            The message to edit.
        content : `None`, `str`, ``EmbedBase``, `Any`, Optional
            The new content of the message.
            
            If given as `str` then the message's content will be edited with it. If given as any non ``EmbedBase``
            instance, then it will be cased to string first.
            
            If given as ``EmbedBase``, then the message's embeds will be edited with it.
        embed : `None`, ``EmbedBase``, `list` of ``EmbedBase``, Optional (Keyword only)
            The new embedded content of the message. By passing it as `None`, you can remove the old.
            
            If `embed` and `content` parameters are both given as  ``EmbedBase``, then `AssertionError` is
            raised.
        
        file : `None`, `Any`, Optional (Keyword only)
            A file or files to send. Check ``create_file_form`` for details.
        
        allowed_mentions : `None`,  `str`, ``UserBase``, ``Role``, `list` of (`str`, ``UserBase``, ``Role`` )
                , Optional (Keyword only)
            Which user or role can the message ping (or everyone). Check ``parse_allowed_mentions``
            for details.
        components : `None`, ``ComponentBase``, (`tuple`, `list`) of (``ComponentBase``, (`tuple`, `list`) of
                ``ComponentBase``), Optional (Keyword only)
            Components attached to the message.
            
            Pass it as `None` remove the actual ones.
        
        suppress : `bool`, Optional (Keyword only)
            Whether the message's embeds should be suppressed or unsuppressed.
            
            Deprecated, please use `suppress_embeds` parameter instead.
        
        suppress_embeds : `bool`, Optional (Keyword only)
            Whether the message's embeds should be suppressed or unsuppressed.
        
        Raises
        ------
        TypeError
            - If `embed` was not given neither as ``EmbedBase`` nor as `list`, `tuple` of ``EmbedBase``-s.
            - If `allowed_mentions` contains an element of invalid type.
            - `content` parameter was given as ``EmbedBase``, meanwhile `embed` parameter was given as well.
            - If `message`'s type is incorrect.
            - If `components` type is incorrect.
        ValueError
            - If `allowed_mentions` contains an element of invalid type.
            - If more than `10` files would be sent.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            - If `suppress_embeds` was not given as `bool`.
            - If `embed` contains a non ``EmbedBase`` element.
            - If both `content` and `embed` fields are embeds.
        
        See Also
        --------
        - ``.message_suppress_embeds`` : For suppressing only the embeds of the message.
        - ``.webhook_message_edit`` : Editing messages sent by webhooks.
        
        Notes
        -----
        Do not updates he given message object, so dispatch event events can still calculate differences when received.
        """
        message, channel_id, message_id = get_message_and_channel_id_and_message_id(message)
        
        content, embed = validate_content_and_embed(content, embed, True)
        
        components = get_components_data(components, True)
        
        if (suppress is not ...):
            warnings.warn(
                (
                    f'`{self.__class__.__name__}.message_edit`\'s `suppress` parameter is deprecated, and '
                    f'will be removed in 2022 May. Please use `suppress_embeds` instead.'
                ),
                FutureWarning,
                stacklevel = 2,
            )
            
            suppress_embeds = suppress
        
        if __debug__:
            if (suppress_embeds is not ...) and (not isinstance(suppress_embeds, bool)):
                raise AssertionError(
                    f'`suppress_embeds` can be `bool`, got {suppress_embeds.__class__.__name__}; {suppress_embeds!r}.'
                )
        
        # Build payload
        message_data = {}
        
        if (content is not ...):
            message_data['content'] = content
        
        if (embed is not ...):
            if (embed is not None):
                embed = [embed.to_data() for embed in embed]
            
            message_data['embeds'] = embed
        
        if (allowed_mentions is not ...):
            message_data['allowed_mentions'] = parse_allowed_mentions(allowed_mentions)
        
        if (components is not ...):
            message_data['components'] = components
        
        if (suppress is not ...):
            if message is None:
                flags = 0
            else:
                flags = message.flags
            
            if suppress_embeds:
                flags |= MESSAGE_FLAG_VALUE_SUPPRESS_EMBEDS
            else:
                flags &= ~MESSAGE_FLAG_VALUE_SUPPRESS_EMBEDS
            
            message_data['flags'] = flags
        
        message_data = add_file_to_message_data(message_data, file, True, True)
        if (message_data is None):
            return
        
        await self.http.message_edit(channel_id, message_id, message_data)
    
    
    async def message_delete(self, message, *, reason=None):
        """
        Deletes the given message.
        
        This method is a coroutine.
        
        Parameters
        ----------
        message : ``Message``, ``MessageReference``, ``MessageRepr``, `tuple` (`int`, `int`)
            The message to delete.
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Shows up at the respective guild's audit logs.
        
        Raises
        ------
        TypeError
            If message was not given neither as ``Message``, ``MessageReference``, ``MessageRepr``, neither as
            `tuple` (`int`, `int`).
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Notes
        -----
        The rate limit group is different for own or for messages newer than 2 weeks than for message's of others,
        which are older than 2 weeks.
        """
        
        channel_id, message_id, message = validate_message_to_delete(message)
        
        if (message is None):
            author = None
        else:
            if not message.is_deletable():
                return
            
            author = message.author
        
        if (author is self) or (message_id > int((time_now() - 1209590.) * 1000.-DISCORD_EPOCH) << 22):
            # own or new
            coroutine = self.http.message_delete(channel_id, message_id, reason)
        else:
            coroutine = self.http.message_delete_b2wo(channel_id, message_id, reason)
        
        await coroutine
        # If the coroutine raises, do not switch `message.deleted` to `True`.
        if (message is not None):
            message.deleted = True
    
    
    async def message_delete_multiple(self, messages, *, reason=None):
        """
        Deletes the given messages. The messages can be from different channels as well.
        
        This method is a coroutine.
        
        Parameters
        ----------
        messages : (`list`, `set`, `tuple`) of \
                (``Message``, ``MessageReference``, ``MessageRepr``, `tuple` (`int`, `int`))
            The messages to delete.
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Shows up at the respective guild's audit logs.
        
        Raises
        ------
        TypeError
            If a message was not given neither as ``Message``, ``MessageReference``, ``MessageRepr``, neither as
            `tuple` (`int`, `int`).
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            If `messages` was not given neither as `list`, `set` nor as `tuple`.
        
        Notes
        -----
        This method uses up 3 different rate limit groups parallelly to maximize the deletion speed.
        """
        if __debug__:
            if not isinstance(messages, (list, set, tuple)):
                raise AssertionError(
                    f'`messages` can be `list`, `set`, `tuple`, got {messages.__class__.__name__}; {messages!r}.'
                )
        
        if not messages:
            return
        
        bulk_delete_limit = int((time_now() - 1209600.) * 1000.-DISCORD_EPOCH) << 22 # 2 weeks
        
        by_channel = {}
        
        for message in messages:
            channel_id, message_id, message = validate_message_to_delete(message)
            if (message is not None) and (not message.is_deletable()):
                continue
            
            if message is None:
                own = False
            else:
                if message.author is self:
                    own = True
                else:
                    own = False
            
            try:
                message_group_new, message_group_old, message_group_old_own = by_channel[channel_id]
            except KeyError:
                message_group_new = deque()
                message_group_old = deque()
                message_group_old_own = deque()
                by_channel[channel_id] = (message_group_new, message_group_old, message_group_old_own)
            
            if message_id > bulk_delete_limit:
                message_group_new.append((own, message_id),)
                continue
            
            if own:
                group = message_group_old_own
            else:
                group = message_group_old
            
            group.append(message_id)
            continue
        
        tasks = []
        for channel_id, groups in by_channel.items():
            try:
                channel = CHANNELS[channel_id]
            except KeyError:
                is_private = False
            else:
                if channel.partial:
                    is_private = True
                else:
                    is_private = channel.is_in_group_private()
            
            if is_private:
                function = _message_delete_multiple_private_task
            else:
                function = _message_delete_multiple_task
            
            task = Task(function(self, channel_id, groups, reason), KOKORO)
            tasks.append(task)
        
        await WaitTillAll(tasks, KOKORO)
        
        last_exception = None
        for task in tasks:
            exception = task.exception()
            if exception is None:
                continue
            
            if last_exception is None:
                last_exception = exception
            else:
                if isinstance(exception, ConnectionError):
                    # This is the lowest priority exception, never overwrite older ones.
                    pass
                elif isinstance(exception, DiscordException):
                    # Do overwrite only `ConnectionError`.
                    if isinstance(last_exception, ConnectionError):
                        last_exception = exception
                else:
                    # Do not overwrite same tier exceptions again, only `ConnectionError` and `DiscordException`
                    if isinstance(last_exception, (ConnectionError, DiscordException)):
                        last_exception = exception
            
            task.cancel()
        
        if (last_exception is not None):
            raise last_exception
    
        
    async def message_delete_sequence(self, channel, *, after=None, before=None, limit=None, filter=None, reason=None):
        """
        Deletes messages between an interval determined by `before` and `after`. They can be `int`, or as
        a ``DiscordEntity`` or as a `datetime` object.
        
        If the client has no `manage_messages` permission at the channel, then returns instantly.
        
        This method is a coroutine.
        
        Parameters
        ----------
        channel : ``Channel``
            The channel, where the deletion should take place.
        after : `None`, `int`, ``DiscordEntity``, `datetime` = `None`, Optional (Keyword only)
            The timestamp after the messages were created, which will be deleted.
        before : `None`, `int`, ``DiscordEntity``, `datetime` = `None`, Optional (Keyword only)
            The timestamp before the messages were created, which will be deleted.
        limit : `None`, `int` = `None`, Optional (Keyword only)
            The maximal amount of messages to delete.
        filter : `None`, `callable` = `None`, Optional (Keyword only)
            A callable filter, what should accept a message object as parameter and return either `True`, `False`.
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Shows up at the respective guild's audit logs.
        
        Raises
        ------
        TypeError
            If `after`, `before` was passed with an unexpected type.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            If `channel` was not given as ``Channel``.
        
        Notes
        -----
        This method uses up 4 different rate limit groups parallelly to maximize the request and the deletion speed.
        """
        if __debug__:
            if not (isinstance(channel, Channel) and (channel.is_in_group_messageable() or channel.partial)):
                raise AssertionError(
                    f'`channel` can be a messageable channel, got {channel.__class__.__name__}; {channel!r}.'
                )
        
        # Check permissions
        permissions = channel.cached_permissions_for(self)
        if not permissions & PERMISSION_MASK_MANAGE_MESSAGES:
            return
        
        before = 9223372036854775807 if before is None else log_time_converter(before)
        after = 0 if after is None else log_time_converter(after)
        limit = 9223372036854775807 if limit is None else limit
        
        # Check for reversed intervals
        if before < after:
            return
        
        # Check if we are done already
        if limit <= 0:
            return
        
        message_group_new = deque()
        message_group_old = deque()
        message_group_old_own = deque()
        
        # Check if we can request more messages
        if channel.message_history_reached_end or (not permissions & PERMISSION_MASK_READ_MESSAGE_HISTORY):
            should_request = False
        else:
            should_request = True
        
        last_message_id = before
        
        messages_ = channel.messages
        if (messages_ is not None) and messages_:
            before_index = message_relative_index(messages_, before)
            after_index = message_relative_index(messages_, after)
            if before_index != after_index:
                time_limit = int((time_now() - 1209600.) * 1000.-DISCORD_EPOCH) << 22
                while True:
                    if before_index == after_index:
                        break
                    
                    message_ = messages_[before_index]
                    before_index += 1
                    
                    # Ignore invoking user only messages! Desu!
                    if not message_.is_deletable():
                        continue
                    
                    if (filter is not None):
                        if not filter(message_):
                            continue
                    
                    last_message_id = message_.id
                    own = (message_.author is self)
                    if last_message_id > time_limit:
                        message_group_new.append((own, last_message_id,),)
                    else:
                        if own:
                            group = message_group_old_own
                        else:
                            group = message_group_old
                        group.append(last_message_id)
                    
                    # Check if we reached the limit
                    limit -= 1
                    if limit:
                        continue
                    
                    should_request = False
                    break
        
        tasks = []
        
        get_mass_task = None
        delete_mass_task = None
        delete_new_task = None
        delete_old_task = None
        
        channel_id = channel.id
        
        while True:
            if should_request and (get_mass_task is None):
                request_data = {
                    'limit': 100,
                    'before': last_message_id,
                }
                
                get_mass_task = Task(self.http.message_get_chunk(channel_id, request_data), KOKORO)
                tasks.append(get_mass_task)
            
            if (delete_mass_task is None):
                message_limit = len(message_group_new)
                # If there are more messages, we are waiting for other tasks
                if message_limit:
                    time_limit = int((time_now() - 1209590.) * 1000.-DISCORD_EPOCH) << 22 # 2 weeks -10s
                    collected = 0
                    
                    while True:
                        if collected == message_limit:
                            break
                        
                        if collected == 100:
                            break
                        
                        own, message_id = message_group_new[collected]
                        if message_id < time_limit:
                            break
                        
                        collected += 1
                        continue
                    
                    if collected == 0:
                        pass
                    
                    elif collected == 1:
                        # Delete the message if we don't delete a new message already
                        if (delete_new_task is None):
                            # We collected 1 message -> We cannot use mass delete on this.
                            own, message_id = message_group_new.popleft()
                            delete_new_task = Task(self.http.message_delete(channel_id, message_id, reason),
                                KOKORO)
                            tasks.append(delete_new_task)
                    else:
                        message_ids = []
                        while collected:
                            collected -= 1
                            own, message_id = message_group_new.popleft()
                            message_ids.append(message_id)
                        
                        delete_mass_task = Task(self.http.message_delete_multiple(channel_id, {'messages': message_ids},
                            reason), KOKORO)
                        tasks.append(delete_mass_task)
                    
                    # After we checked what is at this group, lets move the others from it's end, if needed ofc
                    message_limit = len(message_group_new)
                    if message_limit:
                        # time limit -> 2 week
                        time_limit = time_limit - 20971520000
                        
                        while True:
                            # Cannot start at index = len(...), so we instantly do -1
                            message_limit -= 1
                            
                            own, message_id = message_group_new[message_limit]
                            # Check if we should not move -> leave
                            if message_id > time_limit:
                                break
                            
                            del message_group_new[message_limit]
                            if own:
                                group = message_group_old_own
                            else:
                                group = message_group_old
                                
                            group.appendleft(message_id)
                            
                            if message_limit:
                                continue
                            
                            break
            
            if (delete_new_task is None):
                # Check old own messages only, mass delete speed is pretty good by itself.
                if message_group_old_own:
                    message_id = message_group_old_own.popleft()
                    delete_new_task = Task(self.http.message_delete(channel_id, message_id, reason), KOKORO)
                    tasks.append(delete_new_task)
            
            if (delete_old_task is None):
                if message_group_old:
                    message_id = message_group_old.popleft()
                    delete_old_task = Task(self.http.message_delete_b2wo(channel_id, message_id, reason), KOKORO)
                    tasks.append(delete_old_task)
            
            if not tasks:
                # It can happen, that there are no more tasks left, at that case we check if there is more message
                # left. Only at `message_group_new` can be anymore message, because there is a time interval of
                # 10 seconds, what we do not move between categories.
                if not message_group_new:
                    break
                
                # We really have at least 1 message at that interval.
                own, message_id = message_group_new.popleft()
                # We will delete that message with old endpoint if not own, to make sure it will not block the other
                # endpoint for 2 minutes with any chance.
                if own:
                    delete_new_task = Task(self.http.message_delete(channel_id, message_id, reason), KOKORO)
                    task = delete_new_task
                else:
                    delete_old_task = Task(self.http.message_delete_b2wo(channel_id, message_id, reason), KOKORO)
                    task = delete_old_task
                
                tasks.append(task)
            
            done, pending = await WaitTillFirst(tasks, KOKORO)
            
            for task in done:
                tasks.remove(task)
                try:
                    result = task.result()
                except:
                    for task in tasks:
                        task.cancel()
                    raise
                
                if task is get_mass_task:
                    get_mass_task = None
                    
                    received_count = len(result)
                    if received_count < 100:
                        should_request = False
                        
                        # We got 0 messages, move on the next task
                        if received_count == 0:
                            continue
                    
                    # We don't really care about the limit, because we check message id when we delete too.
                    time_limit = int((time_now() - 1209600.) * 1000.-DISCORD_EPOCH) << 22 # 2 weeks
                    
                    for message_data in result:
                        if (filter is None):
                            last_message_id = int(message_data['id'])
    
                            # Did we reach the after limit?
                            if last_message_id < after:
                                should_request = False
                                break
                            
                            # If filter is `None`, we just have to decide, if we were the author or nope.
                            
                            # Try to get user id, first start it with trying to get author data. The default author_id
                            # will be 0, because that's sure not the id of the client.
                            try:
                                author_data = message_data['author']
                            except KeyError:
                                author_id = 0
                            else:
                                # If we have author data, lets select the user's data from it
                                try:
                                    user_data = author_data['user']
                                except KeyError:
                                    user_data = author_data
                                
                                try:
                                    author_id = user_data['id']
                                except KeyError:
                                    author_id = 0
                                else:
                                    author_id = int(author_id)
                        else:
                            message_ = Message(message_data)
                            last_message_id = message_.id
                            
                            # Did we reach the after limit?
                            if last_message_id < after:
                                should_request = False
                                break
                            
                            if not filter(message_):
                                continue
                            
                            author_id = message_.author.id
                        
                        own = (author_id == self.id)
                        
                        if last_message_id > time_limit:
                            message_group_new.append((own, last_message_id,),)
                        else:
                            if own:
                                group = message_group_old_own
                            else:
                                group = message_group_old
                            
                            group.append(last_message_id)
                        
                        # Did we reach the amount limit?
                        limit -= 1
                        if limit:
                            continue
                        
                        should_request = False
                        break
                
                if task is delete_mass_task:
                    delete_mass_task = None
                    continue
                
                if task is delete_new_task:
                    delete_new_task = None
                    continue
                
                if task is delete_old_task:
                    delete_old_task = None
                    continue
                 
                # Should not happen
                continue
    
    
    async def multi_client_message_delete_sequence(
        self, channel, *, after=None, before=None, limit=None, filter=None, reason=None
    ):
        """
        Deletes messages between an interval determined by `before` and `after`. They can be `int`, or as
        a ``DiscordEntity`` or as a `datetime` object.
        
        Not like ``.message_delete_sequence``, this method uses up all he clients at the respective channel to delete
        messages an not only the one from what it was called from.
        
        If non of the clients have `manage_messages` permission, then returns instantly.
        
        This method is a coroutine.
        
        Parameters
        ----------
        channel : ``Channel``
            The channel, where the deletion should take place.
        after : `None`, `int`, ``DiscordEntity``, `datetime` = `None`, Optional (Keyword only)
            The timestamp after the messages were created, which will be deleted.
        before : `None`, `int`, ``DiscordEntity``, `datetime` = `None`, Optional (Keyword only)
            The timestamp before the messages were created, which will be deleted.
        limit : `None`, `int` = `None`, Optional (Keyword only)
            The maximal amount of messages to delete.
        filter : `None`, `callable` = `None`, Optional (Keyword only)
            A callable filter, what should accept a message object as parameter and return either `True`, `False`.
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Shows up at the respective guild's audit logs.
        
        Raises
        ------
        TypeError
            If `after`, `before` was passed with an unexpected type.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            If `channel` is not ``Channel``.
        
        Notes
        -----
        This method uses up to 4 different endpoint groups too as ``.message_delete_sequence``, but tries to
        parallelize the them between more clients as well.
        """
        if __debug__:
            if not (isinstance(channel, Channel) and (channel.is_in_group_messageable() or channel.partial)):
                raise AssertionError(
                    f'`channel` can be a messageable channel, got {channel.__class__.__name__}; {channel!r}.'
                )
        
        # Check permissions
        sharders = []
        
        for client in channel.clients:
            sharder = MultiClientMessageDeleteSequenceSharder(client, channel)
            if sharder is None:
                continue
            
            sharders.append(sharder)
        
        if not sharders:
            return
        
        for sharder in sharders:
            if sharder.can_manage_messages:
                break
        else:
            return
        
        before = 9223372036854775807 if before is None else log_time_converter(before)
        after = 0 if after is None else log_time_converter(after)
        limit = 9223372036854775807 if limit is None else limit
        
        # Check for reversed intervals
        if before < after:
            return
        
        # Check if we are done already
        if limit <= 0:
            return
        
        message_group_new = deque()
        message_group_old = deque()
        message_group_old_own = deque()
        
        # Check if we can request more messages
        if channel.message_history_reached_end:
            should_request = False
        else:
            for sharder in sharders:
                if sharder.can_read_message_history:
                    should_request = True
                    break
            else:
                should_request = False
        
        last_message_id = before
        
        is_own_getter = {}
        for index in range(len(sharders)):
            is_own_getter[sharders[index].client.id] = index
        
        messages_ = channel.messages
        if (messages_ is not None) and messages_:
            before_index = message_relative_index(messages_, before)
            after_index = message_relative_index(messages_, after)
            if before_index != after_index:
                time_limit = int((time_now() - 1209600.) * 1000.-DISCORD_EPOCH) << 22
                while True:
                    if before_index == after_index:
                        break
                    
                    message_ = messages_[before_index]
                    before_index += 1
                    
                    if not message_.is_deletable():
                        continue
                    
                    if (filter is not None):
                        if not filter(message_):
                            continue
                    
                    last_message_id = message_.id
                    who_s = is_own_getter.get(message_.author.id, -1)
                    if last_message_id > time_limit:
                        message_group_new.append((who_s, last_message_id,),)
                    else:
                        if who_s == -1:
                            message_group_old.append(last_message_id)
                        else:
                            message_group_old_own.append((who_s, last_message_id,),)
                    
                    # Check if we reached the limit
                    limit -= 1
                    if limit:
                        continue
                    
                    should_request = False
                    break
        
        tasks = []
        # Handle requesting together, since we need to know, till where the last request yielded.
        get_mass_task = None
        # Loop the sharders when requesting, so rate limits are used up.
        get_mass_task_next = 0
        
        channel_id = channel.id
        
        while True:
            if should_request and (get_mass_task is None):
                # Will break since `should_request` is set to `True` only if at least of the sharders have
                # `read_message_history` permission
                while True:
                    if get_mass_task_next >= len(sharders):
                        get_mass_task_next = 0
                    
                    sharder = sharders[get_mass_task_next]
                    if sharder.can_read_message_history:
                        break
                    
                    get_mass_task_next += 1
                    continue
                
                request_data = {
                    'limit': 100,
                    'before': last_message_id,
                }
                
                get_mass_task = Task(sharder.client.http.message_get_chunk(channel_id, request_data), KOKORO)
                tasks.append(get_mass_task)
            
            for sharder in sharders:
                if (sharder.can_manage_messages) and (sharder.delete_mass_task is None):
                    message_limit = len(message_group_new)
                    # If there are more messages, we are waiting for other tasks
                    if message_limit:
                        time_limit = int((time_now() - 1209590.) * 1000.-DISCORD_EPOCH) << 22 # 2 weeks -10s
                        collected = 0
                        
                        while True:
                            if collected == message_limit:
                                break
                            
                            if collected == 100:
                                break
                            
                            who_s, message_id = message_group_new[collected]
                            if message_id < time_limit:
                                break
                            
                            collected += 1
                            continue
                        
                        if collected == 0:
                            pass
                        
                        elif collected == 1:
                            # Delete the message if we don't delete a new message already
                            for sub_sharder in sharders:
                                if (sub_sharder.can_manage_messages) and (sharder.delete_new_task is None):
                                    # We collected 1 message -> We cannot use mass delete on this.
                                    who_s, message_id = message_group_new.popleft()
                                    delete_new_task = Task(sub_sharder.client.http.message_delete(channel_id,
                                        message_id, reason=reason), KOKORO)
                                    sub_sharder.delete_new_task = delete_new_task
                                    tasks.append(delete_new_task)
                                    break
                        else:
                            message_ids = []
                            while collected:
                                collected -= 1
                                who_s, message_id = message_group_new.popleft()
                                message_ids.append(message_id)
                            
                            delete_mass_task = Task(sharder.client.http.message_delete_multiple(channel_id,
                                {'messages': message_ids}, reason), KOKORO)
                            sharder.delete_mass_task = delete_mass_task
                            tasks.append(delete_mass_task)
                        
                        # After we checked what is at this group, lets move the others from it's end, if needed ofc
                        message_limit = len(message_group_new)
                        if message_limit:
                            # time limit -> 2 week
                            time_limit = time_limit - 20971520000
                            
                            while True:
                                # Cannot start at index = len(...), so we instantly do -1
                                message_limit -= 1
                                
                                who_s, message_id = message_group_new[message_limit]
                                # Check if we should not move -> leave
                                if message_id > time_limit:
                                    break
                                
                                del message_group_new[message_limit]
                                if who_s == -1:
                                    message_group_old.appendleft(message_id)
                                else:
                                    message_group_old_own.appendleft((who_s, message_group_old,),)
                                
                                if message_limit:
                                    continue
                                
                                break
            
            # Check old own messages only, mass delete speed is pretty good by itself.
            if message_group_old_own:
                # Check who's is the last message. And delete with it. These speed is pretty fast.
                # I doubt it needs further speedup, since deleting not own messages are the bottleneck of message
                # deletions.
                who_s, message_id = message_group_old_own[0]
                sharder = sharders[who_s]
                if sharder.delete_new_task is None:
                    del message_group_old_own[0]
                    delete_new_task = Task(sharder.client.http.message_delete(channel_id, message_id, reason), KOKORO)
                    sharder.delete_new_task = delete_new_task
                    tasks.append(delete_new_task)
            
            if message_group_old:
                for sharder in sharders:
                    if (sharder.delete_old_task is None):
                        message_id = message_group_old.popleft()
                        delete_old_task = Task(
                            sharder.client.http.message_delete_b2wo(channel_id, message_id, reason), KOKORO
                        )
                        sharder.delete_old_task = delete_old_task
                        tasks.append(delete_old_task)
                        
                        if not message_group_old:
                            break
            
            if not tasks:
                # It can happen, that there are no more tasks left, at that case we check if there is more message
                # left. Only at `message_group_new` can be anymore message, because there is a time interval of
                # 10 seconds, what we do not move between categories.
                if not message_group_new:
                    break
                
                # We really have at least 1 message at that interval.
                who_s, message_id = message_group_new.popleft()
                # We will delete that message with old endpoint if not own, to make sure it will not block the other
                # endpoint for 2 minutes with any chance.
                if who_s == -1:
                    for sharder in sharders:
                        if sharder.can_manage_messages:
                            task = Task(sharder.client.http.message_delete_b2wo(channel_id, message_id, reason), KOKORO)
                            tasks.append(task)
                            sharder.delete_old_task = task
                            break
                else:
                    sharder = sharders[who_s]
                    task = Task(sharder.client.http.message_delete(channel_id, message_id, reason), KOKORO)
                    tasks.append(task)
                    sharder.delete_new_task = task
            
            
            done, pending = await WaitTillFirst(tasks, KOKORO)
            
            for task in done:
                tasks.remove(task)
                try:
                    result = task.result()
                except:
                    for task in tasks:
                        task.cancel()
                    raise
                
                if task is get_mass_task:
                    get_mass_task = None
                    
                    received_count = len(result)
                    if received_count < 100:
                        should_request = False
                        
                        # We got 0 messages, move on the next task
                        if received_count == 0:
                            continue
                    
                    # We don't really care about the limit, because we check message id when we delete too.
                    time_limit = int((time_now() - 1209600.) * 1000.-DISCORD_EPOCH) << 22 # 2 weeks
                    
                    for message_data in result:
                        if (filter is None):
                            last_message_id = int(message_data['id'])
                            
                            # Did we reach the after limit?
                            if last_message_id < after:
                                should_request = False
                                break
                            
                            # If filter is `None`, we just have to decide, if we were the author or nope.
                            
                            # Try to get user id, first start it with trying to get author data. The default author_id
                            # will be 0, because that's sure not the id of the client.
                            try:
                                author_data = message_data['author']
                            except KeyError:
                                author_id = 0
                            else:
                                # If we have author data, lets select the user's data from it
                                try:
                                    user_data = author_data['user']
                                except KeyError:
                                    user_data = author_data
                                
                                try:
                                    author_id = user_data['id']
                                except KeyError:
                                    author_id = 0
                                else:
                                    author_id = int(author_id)
                        else:
                            message_ = Message(message_data)
                            last_message_id = message_.id
                            
                            # Did we reach the after limit?
                            if last_message_id < after:
                                should_request = False
                                break
                            
                            if not filter(message_):
                                continue
                            
                            author_id = message_.author.id
                        
                        who_s = is_own_getter.get(author_id, -1)
                        
                        if last_message_id > time_limit:
                            message_group_new.append((who_s, last_message_id,),)
                        else:
                            if who_s == -1:
                                message_group_old.append(last_message_id)
                            else:
                                message_group_old_own.append((who_s, last_message_id,),)
                        
                        # Did we reach the amount limit?
                        limit -= 1
                        if limit:
                            continue
                        
                        should_request = False
                        break
                
                for sharder in sharders:
                    if task is sharder.delete_mass_task:
                        sharder.delete_mass_task = None
                        break
                    
                    if task is sharder.delete_new_task:
                        sharder.delete_new_task = None
                        break
                    
                    if task is sharder.delete_old_task:
                        sharder.delete_old_task = None
                        break
                
                # Else case should happen.
                continue
    
    
    async def message_suppress_embeds(self, message, suppress_embeds=True):
        """
        Suppresses or unsuppressed the given message's embeds.
        
        This method is a coroutine.
        
        Parameters
        ----------
        message : ``Message``, ``MessageRepr``, ``MessageReference``, `tuple` (`int`, `int`)
            The message, what's embeds will be (un)suppressed.
        suppress_embeds : `bool` = `True`, Optional
            Whether the message's embeds would be suppressed or unsuppressed.
        
        Raises
        ------
        TypeError
            If `message` was not given neither as ``Message``, ``MessageRepr`, ``MessageReference`` neither as `tuple`
            (`int`, `int`) instance.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            If `suppress` was not given as `bool`.
        """
        message, channel_id, message_id = get_message_and_channel_id_and_message_id(message)
        
        if __debug__:
            if not isinstance(suppress_embeds, bool):
                raise AssertionError(
                    f'`suppress_embeds` can be `bool`, got {suppress_embeds.__class__.__name__}; {suppress_embeds!r}.'
                )
        
        if message is None:
            flags = 0
        else:
            flags = message.flags
        
        if suppress_embeds:
            flags |= MESSAGE_FLAG_VALUE_SUPPRESS_EMBEDS
        else:
            flags &= ~MESSAGE_FLAG_VALUE_SUPPRESS_EMBEDS
        
        await self.http.message_edit(channel_id, message_id, {'flags': flags})
    
    
    async def message_crosspost(self, message):
        """
        Crossposts the given message. The message's channel must be an announcements (type 5) channel.
        
        This method is a coroutine.
        
        Parameters
        ----------
        message : ``Message``, ``MessageRepr``, ``MessageReference``, `tuple` (`int`, `int`)
            The message to crosspost.
        
        Raises
        ------
        TypeError
            If `message` was not given neither as ``Message``, ``MessageRepr`, ``MessageReference`` nor as
            `tuple` (`int`, `int`) instance.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        channel_id, message_id = get_channel_id_and_message_id(message)
        
        await self.http.message_crosspost(channel_id, message_id)
    
    
    async def message_pin(self, message):
        """
        Pins the given message.
        
        This method is a coroutine.
        
        Parameters
        ----------
        message : ``Message``, ``MessageRepr``, ``MessageReference``, `tuple` (`int`, `int`)
            The message to pin.
        
        Raises
        ------
        TypeError
            If `message` was not given neither as ``Message``, ``MessageRepr`, ``MessageReference`` nor as
            `tuple` (`int`, `int`) instance.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        channel_id, message_id = get_channel_id_and_message_id(message)
        
        await self.http.message_pin(channel_id, message_id)
    
    
    async def message_unpin(self, message):
        """
        Unpins the given message.
        
        This method is a coroutine.
        
        Parameters
        ----------
        message : ``Message``, ``MessageRepr``, ``MessageReference``, `tuple` (`int`, `int`)
            The message to unpin.
        
        Raises
        ------
        TypeError
            If `message` was not given neither as ``Message``, ``MessageRepr`, ``MessageReference`` nor as
            `tuple` (`int`, `int`) instance.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        channel_id, message_id = get_channel_id_and_message_id(message)
        
        await self.http.message_unpin(channel_id, message_id)
    
    
    async def channel_pin_get_all(self, channel):
        """
        Returns the pinned messages at the given channel.
        
        This method is a coroutine.
        
        Parameters
        ----------
        channel : ``Channel``, `int`
            The channel from were the pinned messages will be requested.
        
        Returns
        -------
        messages : `list` of ``Message`` objects
            The pinned messages at the given channel.
        
        Raises
        ------
        TypeError
            If `channel` was not given neither as ``Channel`` nor `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        channel_id = get_channel_id(channel, Channel.is_in_group_messageable)
        
        data = await self.http.channel_pin_get_all(channel_id)
        
        return [Message(message_data) for message_data in data]
    
    
    async def _load_messages_till(self, channel, index):
        """
        An internal function to load the messages at the given channel till the given index. Should not be called if
        the channel reached it's message history's end.
        
        This method is a coroutine.
        
        Parameters
        ----------
        channel : ``Channel``
            The channel from where the messages will be requested.
        index : `int`
            Till which index the messages should be requested at the given channel.
        
        Returns
        -------
        result_state : `int`
            can return the following variables describing a state:
            
            +-----------+---------------------------------------------------------------------------+
            | Value     | Description                                                               |
            +-----------+---------------------------------------------------------------------------+
            | 0         | Success.                                                                  |
            +-----------+---------------------------------------------------------------------------+
            | 1         | `index` could not be reached, there is no more messages at the channel.   |
            +-----------+---------------------------------------------------------------------------+
        
        Raises
        ------
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        while True:
            messages = channel.messages
            if messages is None:
                ln = 0
            else:
                ln = len(channel.messages)
            
            load_to = index - ln
            
            # we want to load it till the exact index, so if `load_to` is `0`, that's not enough!
            if load_to < 0:
                result_state = 0
                break
            
            if load_to < 98:
                planned = load_to + 2
            else:
                planned = 100
            
            if ln:
                result = await self.message_get_chunk(channel, planned, before=messages[ln - 1].id + 1)
            else:
                result = await self.message_get_chunk_from_zero(channel, planned)
            
            if len(result) < planned:
                channel.message_history_reached_end = True
                result_state = 1
                break
        
        # Set some collection delay.
        channel._add_message_collection_delay(float(index))
        
        return result_state
    
    
    async def message_get_at_index(self, channel, index):
        """
        Returns the message at the given channel at the specific index. Can be used to load `index` amount of messages
        at the channel.
        
        This method is a coroutine.
        
        Parameters
        ----------
        channel : ``Channel``, `int`.
            The channel from were the messages will be requested.
        index : `int`
            The index of the target message.
        
        Returns
        -------
        message : ``Message`` object
        
        Raises
        ------
        TypeError
            If `channel` was not given neither as ``Channel`` nor `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            - If `index` was not given as `int`.
            - If `index` is out of range [0:].
        """
        if __debug__:
            if not isinstance(index, int):
                raise AssertionError(
                    f'`index` can be `int`, got {index.__class__.__name__}; {index!r}.'
                )
            
            if index < 0:
                raise AssertionError(
                    f'`index` is out from the expected [0:] range, got {index!r}.'
                )
        
        channel, channel_id = get_channel_and_id(channel, Channel.is_in_group_messageable)
        if channel is None:
            messages = await self.message_get_chunk_from_zero(channel_id, min(index + 1, 100))
            
            if messages:
                channel = messages[0].channel
            else:
                raise IndexError(index)
        
        messages = channel.messages
        if (messages is not None) and (index < len(messages)):
            raise IndexError(index)
        
        if channel.message_history_reached_end:
            raise IndexError(index)
        
        if await self._load_messages_till(channel, index):
            raise IndexError(index)
        
        # access it again, because it might be modified
        messages = channel.messages
        if messages is None:
            raise IndexError(index)
        
        return messages[index]
    
    
    async def message_get_all_in_range(self, channel, start=0, end=100):
        """
        Returns a list of the message between the `start` - `end` area. If the client has no permission to request
        messages, or there are no messages at the given area returns an empty list.
        
        This method is a coroutine.
        
        Parameters
        ----------
        channel : ``Channel``, `int`
            The channel from were the messages will be requested.
        start : `int` = `0`, Optional
            The first message's index at the channel to be requested. Defaults to `0`.
        end : `int` = `100`, Optional
            The last message's index at the channel to be requested. Defaults to `100`.
        
        Returns
        -------
        messages : `list` of ``Message`` objects
        
        Raises
        ------
        TypeError
            If `channel` was not given neither as ``Channel`` nor `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            - If `start` was not given as `int`.
            - If `start` is out of range [0:].
            - If `end` was not given as `int`.
            - If `end` is out of range [0:].
        """
        if __debug__:
            if not isinstance(start, int):
                raise AssertionError(
                    f'`start` can be `int`, got {start.__class__.__name__}; {start!r}.'
                )
            
            if start < 0:
                raise AssertionError(
                    f'`start` is out from the expected [0:] range, got {start!r}.'
                )
        
            if not isinstance(end, int):
                raise AssertionError(
                    f'`end` can be `int`, got {end.__class__.__name__}; {end!r}.'
                )
            
            if end < 0:
                raise AssertionError(
                    f'`end` is out from the expected [0:] range, got {end!r}.'
                )
        
        channel, channel_id = get_channel_and_id(channel, Channel.is_in_group_messageable)
        if channel is None:
            messages = await self.message_get_chunk_from_zero(channel_id, min(end + 1, 100))
            
            if messages:
                channel = messages[0].channel
            else:
                return []
        
        if end <= start:
            return []
        
        messages = channel.messages
        if messages is None:
            ln = 0
        else:
            ln = len(messages)
        
        if (end >= ln) and (not channel.message_history_reached_end) and \
               channel.cached_permissions_for(self) & PERMISSION_MASK_READ_MESSAGE_HISTORY:
            
            try:
                await self._load_messages_till(channel, end)
            except DiscordException as err:
                if err.code not in (
                    ERROR_CODES.unknown_message, # message deleted
                    ERROR_CODES.unknown_channel, # message's channel deleted
                    ERROR_CODES.missing_access, # client removed
                    ERROR_CODES.missing_permissions, # permissions changed meanwhile
                    ERROR_CODES.cannot_message_user, # user has dm-s disallowed
                ):
                    raise
        
        result = []
        messages = channel.messages
        if (messages is not None):
            for index in range(start, min(end, len(messages))):
                result.append(messages[index])
        
        return result
    
    
    async def message_iterator(self, channel, *, chunk_size=99):
        """
        Returns an asynchronous message iterator over the given text channel.
        
        This method is a coroutine.
        
        Parameters
        ----------
        channel : ``Channel``  or `int`
            The channel from were the messages will be requested.
        chunk_size : `int` = `99`, Optional (Keyword only)
            The amount of messages to request when the currently loaded history is exhausted. For message chaining
            it is preferably `99`.
        
        Returns
        -------
        message_iterator : ``MessageIterator``
        
        Raises
        ------
        TypeError
            If `channel` was not given neither as ``Channel`` nor `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            - If `chunk_size` was not given as `int`.
            - If `chunk_size` is out of range [1:].
        """
        return await MessageIterator(self, channel, chunk_size)



__all__ = ('ChannelMetadataPrivateGroup',)


from scarletio import copy_docs

from ...bases import ICON_TYPE_NONE, IconSlot, Slotted
from ...http import urls as module_urls
from ...permission.permission import PERMISSION_GROUP, PERMISSION_GROUP_OWNER, PERMISSION_NONE
from ...preconverters import preconvert_snowflake, preconvert_str

from .. import channel_types as CHANNEL_TYPES

from .private_base import ChannelMetadataPrivateBase


class ChannelMetadataPrivateGroup(ChannelMetadataPrivateBase, metaclass=Slotted):
    """
    Channel metadata for private channels.
    
    Attributes
    ----------
    users : `list` of ``ClientUserBase``
        The users in the channel.
    icon_hash : `int`
        The channel's icon's hash in `uint128`.
    icon_type : ``iconType``
        The channel's icon's type.
    name : `str`
        The channel's display name. Can be empty string if the channel has no name.
    owner_id : `int`
        The group channel's owner's id.
    
    Class Attributes
    ----------------
    type : `int` = `CHANNEL_TYPES.private_group`
        The channel's type.
    """
    __slots__ = ('name', 'owner_id')
    
    icon = IconSlot('icon', 'icon', module_urls.channel_group_icon_url, module_urls.channel_group_icon_url_as)
    
    type = CHANNEL_TYPES.private_group
    
    
    @copy_docs(ChannelMetadataPrivateBase._created)
    def _created(self, channel_entity, client):
        if (client is not None):
            client.group_channels[channel_entity.id] = channel_entity
    
        
    @copy_docs(ChannelMetadataPrivateBase._delete)
    def _delete(self, channel_entity, client):
        if (client is not None):
            try:
                del client.group_channels[channel_entity.id]
            except KeyError:
                pass
    
    
    @copy_docs(ChannelMetadataPrivateBase._compare_attributes_to)
    def _compare_attributes_to(self, other):
        if not ChannelMetadataPrivateBase._compare_attributes_to(self, other):
            return False
        
        if self.name != other.name:
            return False
        
        if self.owner_id != other.owner_id:
            return False
        
        if self.icon_hash != other.icon_hash:
            return False
        
        if self.icon_type is not other.icon_type:
            return False
        
        return True
    
    
    @copy_docs(ChannelMetadataPrivateBase._get_processed_name)
    def _get_processed_name(self):
        name = self.name
        if name:
            return name
        
        users = self.users
        if users:
            return ', '.join([user.name for user in users])
        
        return 'Unnamed'
    
    
    @copy_docs(ChannelMetadataPrivateBase._update_attributes)
    def _update_attributes(self, data):
        ChannelMetadataPrivateBase._update_attributes(self, data)
        
        name = data.get('name', None)
        if name is None:
            name = ''
        self.name = name
        
        self.owner_id = int(data['owner_id'])
        
        self._set_icon(data)
    
    
    @copy_docs(ChannelMetadataPrivateBase._difference_update_attributes)
    def _difference_update_attributes(self, data):
        old_attributes = ChannelMetadataPrivateBase._difference_update_attributes(self, data)
        
        name = data.get('name', None)
        if name is None:
            name = ''
        if self.name != name:
            old_attributes['name'] = self.name
            self.name = name
        
        self._update_icon(data, old_attributes)
        
        owner_id = int(data['owner_id'])
        if self.owner_id != owner_id:
            old_attributes['owner_id'] = self.owner_id
            self.owner_id = owner_id
        
        return old_attributes

    
    @copy_docs(ChannelMetadataPrivateBase._get_permissions_for)
    def _get_permissions_for(self, channel_entity, user):
        if self.owner_id == user.id:
            return PERMISSION_GROUP_OWNER
        elif user in self.users:
            return PERMISSION_GROUP
        else:
            return PERMISSION_NONE
    
    
    @classmethod
    @copy_docs(ChannelMetadataPrivateBase._from_partial_data)
    def _from_partial_data(cls, data):
        self = super(ChannelMetadataPrivateGroup, cls)._from_partial_data(data)
        
        if (data is not None):
            name = data.get('name', None)
            if (name is not None):
                self.name = name
        
        return self
    
    
    @classmethod
    @copy_docs(ChannelMetadataPrivateBase._create_empty)
    def _create_empty(cls):
        self = super(ChannelMetadataPrivateGroup, cls)._create_empty()
        
        self.name = None
        self.owner_id = 0
        
        self.icon_hash = 0
        self.icon_type = ICON_TYPE_NONE
        
        return self
    
    
    @classmethod
    @copy_docs(ChannelMetadataPrivateBase._precreate)
    def _precreate(cls, keyword_parameters):
        self = super(ChannelMetadataPrivateGroup, cls)._precreate(keyword_parameters)
        
        try:
            name = keyword_parameters.pop('name')
        except KeyError:
            pass
        else:
            name = preconvert_str(name, 'name', 2, 100)
            self.name = name
        
        
        processable = []
        cls.icon.preconvert(keyword_parameters, processable)
        if processable:
            for item in processable:
                setattr(self, *item)
        processable = None
        
        
        try:
            owner_id = keyword_parameters.pop('owner_id')
        except KeyError:
            pass
        else:
            owner_id = preconvert_snowflake(owner_id, 'owner_id')
            self.owner_id = owner_id
        
        return self
    
    
    @copy_docs(ChannelMetadataPrivateBase._to_data)
    def _to_data(self):
        data = ChannelMetadataPrivateBase.to_data(self)
        
        # name
        name = self.name
        if not name:
            name = None
        data['name'] = name
        
        # owner_id
        data['owner_id'] = str(self.owner_id)
        
        # icon
        data['icon'] = self.icon.as_base16_hash
        
        return self



__all__ = ('ChannelMetadataGuildForum',)


from scarletio import copy_docs

from ...permission import Permission
from ...permission.permission import PERMISSION_MASK_VIEW_CHANNEL, PERMISSION_NONE, PERMISSION_THREAD_AND_VOICE_DENY
from ...preconverters import preconvert_int, preconvert_int_options, preconvert_str

from .. import channel_types as CHANNEL_TYPES
from ..constants import AUTO_ARCHIVE_DEFAULT, AUTO_ARCHIVE_OPTIONS

from .guild_main_base import ChannelMetadataGuildMainBase


class ChannelMetadataGuildForum(ChannelMetadataGuildMainBase):
    """
    Guild forum channel metadata.
    
    Attributes
    ----------
    _permission_cache : `None`, `dict` of (`int`, ``Permission``) items
        A `user_id` to ``Permission`` relation mapping for caching permissions. Defaults to `None`.
    parent_id : `int`
        The channel's parent's identifier.
    name : `str`
        The channel's name.
    permission_overwrites : `dict` of (`int`, ``PermissionOverwrite``) items
        The channel's permission overwrites.
    position : `int`
        The channel's position.
    default_auto_archive_after : `int`
        The default duration (in seconds) for newly created threads to automatically archive the themselves. Defaults
        to `3600`. Can be one of: `3600`, `86400`, `259200`, `604800`.
    slowmode : `int`
        The amount of time in seconds what a user needs to wait between it's each message. Bots and user accounts with
        `manage_messages`, `manage_channel` permissions are unaffected.
    topic : `None`, `str`
        The channel's topic.
    
    Class Attributes
    ----------------
    type : `int` = `CHANNEL_TYPES.guild_forum`
        The channel's type.
    order_group: `int` = `0`
        The channel's order group used when sorting channels.
    """
    __slots__ = ('default_auto_archive_after', 'slowmode', 'topic',)
    
    type = CHANNEL_TYPES.guild_forum
    
    @copy_docs(ChannelMetadataGuildMainBase._compare_attributes_to)
    def _compare_attributes_to(self, other):
        if not ChannelMetadataGuildMainBase._compare_attributes_to(self, other):
            return False
        
        if self.default_auto_archive_after != other.default_auto_archive_after:
            return False
        
        if self.slowmode != other.slowmode:
            return False
        
        if self.topic != other.topic:
            return False
        
        return True
    
    
    @copy_docs(ChannelMetadataGuildMainBase._get_display_name)
    def _get_display_name(self):
        return self.name.upper()
    
    
    @classmethod
    @copy_docs(ChannelMetadataGuildMainBase._create_empty)
    def _create_empty(cls):
        self = super(ChannelMetadataGuildForum, cls)._create_empty()
        
        self.default_auto_archive_after = AUTO_ARCHIVE_DEFAULT
        self.slowmode = 0
        self.topic = None
        
        return self
    
    
    @copy_docs(ChannelMetadataGuildMainBase._update_attributes)
    def _update_attributes(self, data):
        ChannelMetadataGuildMainBase._update_attributes(self, data)
        
        default_auto_archive_after = data.get('default_auto_archive_duration', None)
        if default_auto_archive_after is None:
            default_auto_archive_after = AUTO_ARCHIVE_DEFAULT
        else:
            default_auto_archive_after *= 60
        self.default_auto_archive_after = default_auto_archive_after
        
        slowmode = data.get('rate_limit_per_user', None)
        if slowmode is None:
            slowmode = 0
        self.slowmode = slowmode
        
        self.topic = data.get('topic', None)
    
    
    @copy_docs(ChannelMetadataGuildMainBase._difference_update_attributes)
    def _difference_update_attributes(self, data):
        old_attributes = ChannelMetadataGuildMainBase._difference_update_attributes(self, data)
        
        
        default_auto_archive_after = data.get('default_auto_archive_duration', None)
        if default_auto_archive_after is None:
            default_auto_archive_after = AUTO_ARCHIVE_DEFAULT
        else:
            default_auto_archive_after *= 60
        if self.default_auto_archive_after != default_auto_archive_after:
            old_attributes['default_auto_archive_after'] = self.default_auto_archive_after
            self.default_auto_archive_after = default_auto_archive_after
        
        
        
        slowmode = data.get('rate_limit_per_user', None)
        if slowmode is None:
            slowmode = 0
        if self.slowmode != slowmode:
            old_attributes['slowmode'] = self.slowmode
            self.slowmode = slowmode
        
        
        topic = data.get('topic', None)
        if self.topic != topic:
            old_attributes['topic'] = self.topic
            self.topic = topic
        
        
        return old_attributes
    
    
    @copy_docs(ChannelMetadataGuildMainBase._get_permissions_for)
    def _get_permissions_for(self, channel_entity, user):
        result = self._get_base_permissions_for(channel_entity, user)
        if not result & PERMISSION_MASK_VIEW_CHANNEL:
            return PERMISSION_NONE
        
        # forum channels do not have thread and voice related permissions
        result &= PERMISSION_THREAD_AND_VOICE_DENY
        
        return Permission(result)
    
    
    @copy_docs(ChannelMetadataGuildMainBase._get_permissions_for_roles)
    def _get_permissions_for_roles(self, channel_entity, roles):
        result = self._get_base_permissions_for_roles(channel_entity, roles)
        if not result & PERMISSION_MASK_VIEW_CHANNEL:
            return PERMISSION_NONE
        
        # forum channels do not have thread and voice related permissions
        result &= PERMISSION_THREAD_AND_VOICE_DENY
        return Permission(result)
    
    
    @classmethod
    @copy_docs(ChannelMetadataGuildMainBase._precreate)
    def _precreate(cls, keyword_parameters):
        self = super(ChannelMetadataGuildForum, cls)._precreate(keyword_parameters)
        
        try:
            default_auto_archive_after = keyword_parameters.pop('default_auto_archive_duration')
        except KeyError:
            pass
        else:
            default_auto_archive_after = preconvert_int_options(
                default_auto_archive_after,
                'default_auto_archive_after',
                AUTO_ARCHIVE_OPTIONS,
            )
            
            self.default_auto_archive_after = default_auto_archive_after
        
        
        try:
            slowmode = keyword_parameters.pop('slowmode')
        except KeyError:
            pass
        else:
            slowmode = preconvert_int(slowmode, 'slowmode', 0, 21600)
            self.slowmode = slowmode
        
        
        try:
            topic = keyword_parameters.pop('topic')
        except KeyError:
            pass
        else:
            if (topic is not None):
                topic = preconvert_str(topic, 'topic', 0, 1024)
                if topic:
                    self.topic = topic
        
        
        return self
    
    
    @copy_docs(ChannelMetadataGuildMainBase._to_data)
    def _to_data(self):
        data = ChannelMetadataGuildMainBase._to_data(self)
        
        # default_auto_archive_duration
        data['default_auto_archive_duration'] = self.default_auto_archive_after // 60
        
        # slowmode
        slowmode = self.slowmode
        if slowmode:
            data['rate_limit_per_user'] = slowmode
        
        # topic
        data['topic'] = self.topic
        
        return data



__all__ = ()

from scarletio import Task, WaitTillAll

from ...discord.core import KOKORO

from .constants import LAVALINK_KEY_SESSION_ID, LAVALINK_KEY_VOICE_SERVER_UPDATE_EVENT


async def handle_voice_client_join(client, voice_state):
    """
    Called when the client joins a voice channel.
    
    This function is a coroutine.
    
    Parameters
    ----------
    client : ``Client``
        The respective client.
    voice_state : ``VoiceState``
        The client's voice state.
    """
    guild_id = voice_state.guild_id
    
    try:
        solar_player = client.solarlink.players[guild_id]
    except KeyError:
        return
    
    solar_player.channel_id = voice_state.channel_id
    
    forward_data = solar_player._forward_data
    if forward_data is None:
        forward_data = {}
        solar_player._forward_data = forward_data
    
    forward_data[LAVALINK_KEY_SESSION_ID] = voice_state.session_id
    
    await solar_player._voice_update()


async def handle_voice_client_move(client, voice_state, old_channel_id):
    """
    Called when the client moves between voice channels.
    
    This function is a coroutine.
    
    Parameters
    ----------
    client : ``Client``
        The respective client.
    voice_state : ``VoiceState``
        The client's voice state.
    old_channel_id : `int`
        The channel's identifier from where there client moved.
    """
    guild_id = voice_state.guild_id
    
    try:
        solar_player = client.solarlink.players[guild_id]
    except KeyError:
        return
    
    solar_player.channel_id = voice_state.channel_id


async def handle_voice_client_update(client, voice_state, old_attributes):
    """
    Called when the client's voice state is edited.
    
    This function is a coroutine.
    
    Parameters
    ----------
    client : ``Client``
        The respective client.
    voice_state : ``VoiceState``
        The client's voice state.
    old_attributes : `dict` of (`str`, `Any`) items
        The voice state's old, changed attributes in `attribute-name` - `old-value` relation.
    """
    guild_id = voice_state.guild_id
    
    try:
        solar_player = client.solarlink.players[guild_id]
    except KeyError:
        return
    
    if 'session_id' not in old_attributes:
        return
    
    forward_data = solar_player._forward_data
    if forward_data is None:
        forward_data = {}
        solar_player._forward_data = forward_data
        forward = False
    else:
        forward = True
        
    forward_data[LAVALINK_KEY_SESSION_ID] = voice_state.session_id
    
    if not forward:
        return
    
    await solar_player._voice_update()


async def handle_voice_client_leave(client, voice_state, old_channel_id):
    """
    Called when the client leaves from a voice channel.
    
    This function is a coroutine.
    
    Parameters
    ----------
    client : ``Client``
        The respective client.
    voice_state : ``VoiceState``
        The client's voice state.
    old_channel_id : `int`
        The channel's identifier from where there client left.
    """
    guild_id = voice_state.guild_id
    
    try:
        solar_player = client.solarlink.players[guild_id]
    except KeyError:
        return
    
    solar_player.channel_id = 0
    await solar_player.disconnect()


async def handle_voice_server_update(client, event):
    """
    Called when a guild's voice server, or when the client joins a voice channel.
    
    This function is a coroutine.
    
    Parameters
    ----------
    client : ``Client``
        The respective client.
    event : ``VoiceServerUpdateEvent``
        The event.
    """
    guild_id = event.guild_id
    
    try:
        solar_player = client.solarlink.players[guild_id]
    except KeyError:
        return
    
    forward_data = solar_player._forward_data
    if forward_data is None:
        forward_data = {}
        solar_player._forward_data = forward_data
    
    forward_data[LAVALINK_KEY_VOICE_SERVER_UPDATE_EVENT] = {
        'guild_id': str(event.guild_id),
        'endpoint': event.endpoint,
        'token': event.token,
    }
    
    await solar_player._voice_update()


async def handle_voice_client_ghost(client, voice_state):
    """
    Called when a ghost voice client is present, when logging into discord.
    
    This function is a coroutine.
    
    Parameters
    ----------
    client : ``Client``
        The respective client.
    voice_state : ``VoiceState``
        The client's voice state.
    """
    player = await client.solarlink.join_voice((voice_state.guild_id, voice_state.channel_id))
    await player.disconnect()


async def handle_voice_client_shutdown(client):
    """
    Called when the client logs out.
    
    This function is a coroutine.
    """
    tasks = []
    for player in client.solarlink.players.values():
        task = Task(player.disconnect(), KOKORO)
        tasks.append(task)
    
    task = None
    await WaitTillAll(tasks, KOKORO)
    
    tasks = []
    for node in client.solarlink.nodes:
        task = Task(node.close(), KOKORO)
        tasks.append(task)
    
    task = None
    await WaitTillAll(tasks, KOKORO)



__all__ = ('maybe_find_commands',)

import sys
from os import listdir as list_directory
from os.path import isdir as is_directory, isfile as is_file, join as join_paths

from scarletio import render_exception_into

from .constants import COMMAND_DIRECTORY, COMMAND_IMPORT_ROUTE, PYTHON_FILE_POSTFIX_NAMES
from .external import get_external_command_routes


def _ignore_import_frame(file_name, name, line_number, line):
    """
    Ignores the frame where the file is imported.
    
    Parameters
    ----------
    file_name : `str`
        The frame's respective file's name.
    name : `str`
        The frame's respective function's name.
    line_number : `int`
        The line's index where the exception occurred.
    line : `str`
        The frame's respective stripped line.
    
    Returns
    -------
    should_show_frame : `bool`
        Whether the frame should be shown.
    """
    should_show_frame = True
    
    if file_name == __file__:
        if name == '_import_route':
            if line == '__import__(import_route)':
                should_show_frame = False
    
    return should_show_frame


def _import_route(import_route):
    """
    Imports the given route.
    
    Parameters
    ----------
    import_route : `str`
        The import route to import.
    
    Returns
    -------
    success : `bool`
    """
    try:
        __import__(import_route)
    except (KeyboardInterrupt, SystemExit):
        raise
    
    except BaseException as err:
        sys.stderr.write(
            ''.join(
                render_exception_into(
                    err,
                    [
                        'Exception occurred while importing "',
                        import_route,
                        '".\n'
                    ],
                    filter = _ignore_import_frame,
                ),
            )
        )
        return False
    
    return True


def _find_internal_commands():
    """
    Finds the commands in the commands directory.
    """
    for file_name in list_directory(COMMAND_DIRECTORY):
        path = join_paths(COMMAND_DIRECTORY, file_name)
        
        if is_file(path):
            for postfix in PYTHON_FILE_POSTFIX_NAMES:
                if file_name.endswith(postfix):
                    import_name = file_name[:-len(postfix)]
                    break
            
            else:
                continue
        
        elif is_directory(path):
            if is_file(join_paths(path, '__init__.py')):
                import_name = file_name
            else:
                continue
        
        else:
            continue
        
        import_route = '.'.join([*COMMAND_IMPORT_ROUTE, import_name])
        _import_route(import_route)


def _find_external_commands():
    """
    Will try find external commands in registered external import routes.
    """
    external_import_routes = get_external_command_routes()
    for import_route in external_import_routes:
        _import_route(import_route)


_FIND_COMMANDS_INVOKED = False

def maybe_find_commands():
    """
    Looks up the local commands if not yet looked up.
    """
    global _FIND_COMMANDS_INVOKED
    if not _FIND_COMMANDS_INVOKED:
        try:
            _find_commands()
        finally:
            _FIND_COMMANDS_INVOKED = True


def _find_commands():
    """
    Looks up the local commands.
    """
    _find_internal_commands()
    _find_external_commands()



__all__ = ()

from ...auto_moderation import AutoModerationRuleTriggerType
from ...permission import PermissionOverwriteTargetType


def detail_converter_channel_id(key, value):
    return 'channel_id', int(value)


def detail_converter_amount(key, value):
    return 'amount', int(value)


def detail_converter_delete_message_days_as_duration(key, value):
    return 'delete_message_duration', int(value) * 24 * 60 * 60


def detail_converter_delete_message_duration(key, value):
    return 'delete_message_duration', int(value)


def detail_converter_users_removed(key, value):
    return 'users_removed', int(value)


def detail_converter_message(key, value):
    return 'message_id', int(value)


def detail_converter_permission_overwrite_target_id(key, value):
    return 'target_id', int(value)


def detail_converter_permission_overwrite_target_type(key, value):
    if PermissionOverwriteTargetType.VALUE_TYPE is int:
        value = int(value)
    
    return 'target_type', PermissionOverwriteTargetType.get(value)


def detail_converter_auto_moderation_rule_trigger_type(key, value):
    return key, AutoModerationRuleTriggerType.get(value)


def detail_converter_nothing(key, value):
    return key, value


DETAIL_CONVERTERS = {
    'auto_moderation_rule_name': detail_converter_nothing,
    'auto_moderation_rule_trigger_type': detail_converter_auto_moderation_rule_trigger_type,
    'channel_id': detail_converter_channel_id,
    'count': detail_converter_amount,
    'delete_message_days': detail_converter_delete_message_days_as_duration,
    'delete_message_seconds': detail_converter_delete_message_duration,
    'id': detail_converter_permission_overwrite_target_id,
    'members_removed': detail_converter_users_removed,
    'message_id': detail_converter_message,
    'role_name': detail_converter_nothing,
    'type': detail_converter_permission_overwrite_target_type,
}



__all__ = ('ChannelMetadataPrivateBase',)

from scarletio import copy_docs

from ...user import User

from .base import ChannelMetadataBase


class ChannelMetadataPrivateBase(ChannelMetadataBase):
    """
    Channel metadata for private channels.
    
    Attributes
    ----------
    users : `list` of ``ClientUserBase``
        The users in the channel.
    
    Class Attributes
    ----------------
    type : `int` = `-1`
        The channel's type.
    order_group: `int` = `0`
        The channel's order group used when sorting channels.
    """
    __slots__ = ('users',)
    
    @copy_docs(ChannelMetadataBase.__new__)
    def __new__(cls, data):
        self = ChannelMetadataBase.__new__(cls, data)
        
        users = []
        for user_data in data['recipients']:
            user = User.from_data(user_data)
            users.append(user)
        
        users.sort()
        self.users = users
        
        return self
    
    
    @copy_docs(ChannelMetadataBase._compare_attributes_to)
    def _compare_attributes_to(self, other):
        if self.users != other.users:
            return False
        
        return True
    
    
    @copy_docs(ChannelMetadataBase._get_users)
    def _get_users(self, channel_entity):
        return self.users
    
    
    @copy_docs(ChannelMetadataBase.name)
    def name(self):
        users = self.users
        if users:
            name = f'Direct Message {users[0].full_name} with {users[1].full_name}'
        else:
            name = f'Direct Message (partial)'
        return name
    
    
    @classmethod
    @copy_docs(ChannelMetadataBase._create_empty)
    def _create_empty(cls):
        self = super(ChannelMetadataPrivateBase, cls)._create_empty()
        
        self.users = []
        
        return self
    
    
    @copy_docs(ChannelMetadataBase._to_data)
    def _to_data(self):
        data = ChannelMetadataBase.to_data(self)
        
        # users
        data['recipients'] = [user.to_data() for user in self.users]
        
        return self



__all__ = ()

import reprlib, warnings

from scarletio import Compound, Theory, change_on_switch

from ...bases import maybe_snowflake_pair
from ...color import Color
from ...core import GUILDS, ROLES
from ...emoji import Emoji
from ...guild import create_partial_guild_from_id
from ...http import DiscordHTTPClient, VALID_ICON_MEDIA_TYPES_EXTENDED
from ...permission import Permission
from ...role import Role
from ...utils import get_image_media_type, image_to_base64

from ..functionality_helpers import role_move_key, role_reorder_valid_roles_sort_key
from ..request_helpers import get_guild_and_id, get_guild_id, get_role_guild_id_and_id


class ClientCompoundRoleEndpoints(Compound):
    
    http : DiscordHTTPClient
    
    @Theory
    async def guild_sync(self, guild): ...
    
    
    async def role_edit(
        self, role, *, name=..., color=..., separated=..., mentionable=..., permissions=..., position=..., icon=...,
        reason=None
    ):
        """
        Edits the role with the given parameters.
        
        This method is a coroutine.
        
        Parameters
        ----------
        role : ``Role``, `tuple` (`int`, `int`)
            The role to edit.
        name : `str`, Optional (Keyword only)
            The role's new name. It's length ca be in range [2:32].
        color : ``Color``, `int`, Optional (Keyword only)
            The role's new color.
        separated : `bool`, Optional (Keyword only)
            Whether the users with this role should be shown up as separated from the others.
        mentionable : `bool`, Optional (Keyword only)
            Whether the role should be mentionable.
        permissions : ``Permission``, `int`, Optional (Keyword only)
            The new permission value of the role.
        position : `int`, Optional (Keyword only)
            The role's new position.
        icon : `None`, ``Emoji``, `bytes-like`, Optional (Keyword only)
            The new icon of the role. Can be `'jpg'`, `'png'`, `'webp'`, `'gif'` image's raw data.
            
            Pass it as an ``Emoji`` to set role's icon to it.
            
            Pass it as `None` to remove the role's current icon.
        
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Shows up at the respective guild's audit logs.
        
        Raises
        ------
        TypeError
            - If `role` was not given neither as ``Role`` nor as `tuple` of (`int`, `int`).
            - If `icon` is neither `None`, `bytes-like`.
        ValueError
            - If default role would be moved.
            - If any role would be moved to position `0`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            - If `name` was not given neither as `None` nor `str`.
            - If `name` length is out of range [2:32].
            - If `color` was not given as `None`, ``Color`` nor as other `int`.
            - If `separated` was not given as `None`, nor as `bool`.
            - If `mentionable` was not given as `None`, nor as `bool˛`.
            - If `permissions` was not given as `None`, ``Permission``, neither as other `int`.
            - If `icon` was passed as `bytes-like`, but it's format is not any of the expected formats.
        """
        snowflake_pair = get_role_guild_id_and_id(role)
        if snowflake_pair is None:
            raise TypeError(
                f'`role` can be `{Role.__name__}`, `tuple` (`int`, `int`), got '
                f'{role.__class__.__name__}; {role!r}.'
            )
        
        guild_id, role_id = snowflake_pair
        
        if (position is not None):
            await self.role_move((guild_id, role_id), position, reason=reason)
        
        data = {}
        
        if (name is not ...):
            if __debug__:
                if not isinstance(name, str):
                    raise AssertionError(
                        f'`name` can be `None`, `str`, got {name.__class__.__name__}; {name!r}.'
                    )
                
                name_length = len(name)
                if name_length < 2 or name_length > 32:
                    raise AssertionError(
                        f'`name` length can be in range [2:32], got {name_length!r}; {name!r}.'
                    )
            
            data['name'] = name
        
        if (color is not ...):
            if __debug__:
                if not isinstance(color, int):
                    raise AssertionError(
                        f'`color` can be `None`, `{Color.__name__}`, `int`, got {color.__class__.__name__}.'
                    )
            
            data['color'] = color
        
        if (separated is not ...):
            if __debug__:
                if not isinstance(separated, bool):
                    raise AssertionError(
                        f'`separated` can be `None`, `bool`, got {separated.__class__.__name__}; {separated!r}.'
                    )
            
            data['hoist'] = separated
        
        if (mentionable is not ...):
            if __debug__:
                if not isinstance(mentionable, bool):
                    raise AssertionError(
                        f'`mentionable` can be `None`, `bool`, got {mentionable.__class__.__name__}.'
                    )
            
            data['mentionable'] = mentionable
        
        if (permissions is not ...):
            if __debug__:
                if not isinstance(permissions, int):
                    raise AssertionError(
                        f'`permissions` can be `None`, `{Permission.__name__}`, `int`, got '
                        f'{permissions.__class__.__name__}; {permissions!r}.'
                    )
            
            data['permissions'] = permissions
        
        if (icon is not ...):
            if icon is None:
                data['icon'] = None
            
            elif isinstance(icon, Emoji):
                if icon.is_unicode_emoji():
                    data['unicode_emoji'] = icon.unicode
                elif icon.is_custom_emoji():
                    async with self.http.get(icon.url) as response:
                        icon = await response.read()
                    
                    data['icon'] = image_to_base64(icon)
                
                # no more cases
            
            elif isinstance(icon, (bytes, bytearray, memoryview)):
                if __debug__:
                    media_type = get_image_media_type(icon)
                    
                    if media_type not in VALID_ICON_MEDIA_TYPES_EXTENDED:
                        raise AssertionError(
                            f'Invalid `icon` type for the role: {media_type}; got {reprlib.repr(icon)}.'
                        )
                
                data['icon'] = image_to_base64(icon)
            
            else:
                raise TypeError(
                    f'`icon` can be `None`, `{Emoji.__name__}`, `bytes-like`, got '
                    f'{icon.__class__.__name__}; {reprlib.repr(icon)}.'
                )
        
        
        if data:
            await self.http.role_edit(guild_id, role_id, data, reason)
    
    
    async def role_delete(self, role, *, reason=None):
        """
        Deletes the given role.
        
        This method is a coroutine.
        
        Parameters
        ----------
        role : ``Role``, `tuple` (`int`, `int`)
            The role to delete
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Shows up at the respective guild's audit logs.
        
        Raises
        ------
        TypeError
            If `role` was not given neither as ``Role`` nor as `tuple` of (`int`, `int`).
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        snowflake_pair = get_role_guild_id_and_id(role)
        if snowflake_pair is None:
            raise TypeError(
                f'`role` can be `{Role.__name__}`, `tuple` (`int`, `int`), got {role.__class__.__name__}; {role!r}.'
            )
        
        guild_id, role_id = snowflake_pair
        
        await self.http.role_delete(guild_id, role_id, reason)
    
    
    async def role_create(
        self, guild, *, name=None, permissions=None, color=None, separated=None, mentionable=None, icon=None,
        reason=None
    ):
        """
        Creates a role at the given guild.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild where the role will be created.
        name : `None`, `str` = `None`, Optional (Keyword only)
            The created role's name. It's length can be in range [2:32].
        color : `None`, ``Color``, `int` = `None`, Optional (Keyword only)
            The created role's color.
        separated : `None`, `bool` = `None`, Optional (Keyword only)
            Whether the users with the created role should show up as separated from the others.
        mentionable : `None`, `bool` = `None`, Optional (Keyword only)
            Whether the created role should be mentionable.
        permissions : `None`, ``Permission``, `int` = `None`, Optional (Keyword only)
            The permission value of the created role.
        icon : `None`, ``Emoji``, `bytes-like` = `None`, Optional (Keyword only)
            The icon for the role.
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Shows up at the guild's audit logs.
        
        Raises
        ------
        TypeError
            - If `guild` was not given neither as ``Guild`` nor as `int`.
            - If `icon` is neither `None`, ``Emoji``,  or `bytes-like`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            - If `name` was not given neither as `None` nor `str`.
            - If `name` length is out of range [2:32].
            - If `color` was not given as `None`, ``Color`` nor as other `int`.
            - If `separated` was not given as `None`, nor as `bool`.
            - If `mentionable` was not given as `None`, nor as `bool˛`.
            - If `permissions` was not given as `None`, ``Permission``, neither as other `int`.
            - If `icon` is passed as `bytes-like`, but it's format is not a valid image format.
        """
        guild_id = get_guild_id(guild)
        
        data = {}
        
        if (name is not None):
            if __debug__:
                if not isinstance(name, str):
                    raise AssertionError(
                        f'`name` can be `None`, `str`, got {name.__class__.__name__}; {name!r}.'
                    )
                
                name_length = len(name)
                if name_length < 2 or name_length > 32:
                    raise AssertionError(
                        f'`name` length can be in range [2:32], got {name_length!r}; {name!r}.'
                    )
            
            data['name'] = name
        
        if (permissions is not None):
            if __debug__:
                if not isinstance(color, int):
                    raise AssertionError(
                        f'`permissions` can be `None`, `{Permission.__name__}`, `int`, got '
                        f'{permissions.__class__.__name__}; {permissions!r}.'
                    )
            
            data['permissions'] = permissions
        
        if (color is not None):
            if __debug__:
                if not isinstance(color, int):
                    raise AssertionError(
                        f'`color` can be `None`, `{Color.__name__}`, `int`, got {color.__class__.__name__}; {color!r}.'
                    )
            
            data['color'] = color
        
        if (separated is not None):
            if __debug__:
                if not isinstance(separated, bool):
                    raise AssertionError(
                        f'`separated` can be `None`, `bool`, got {separated.__class__.__name__}; {separated!r}.'
                    )
            
            data['hoist'] = separated
        
        if (mentionable is not None):
            if __debug__:
                if not isinstance(mentionable, bool):
                    raise AssertionError(
                        f'`mentionable` can be `None`, `bool`, got {mentionable.__class__.__name__}; {mentionable!r}.'
                    )
            
            data['mentionable'] = mentionable
        
        if icon is None:
            pass
        
        elif isinstance(icon, Emoji):
            if icon.is_unicode_emoji():
                data['unicode_emoji'] = icon.unicode
            
            elif icon.is_custom_emoji():
                async with self.http.get(icon.url) as response:
                    icon = await response.read()
                
                data['icon'] = image_to_base64(icon)
            
            # no more cases
        
        elif not isinstance(icon, (bytes, bytearray, memoryview)):
            if __debug__:
                media_type = get_image_media_type(icon)
                
                if media_type not in VALID_ICON_MEDIA_TYPES_EXTENDED:
                    raise AssertionError(
                        f'Invalid `icon` type for the role: {media_type}; got {reprlib.repr(icon)}.'
                    )
            
            data['icon'] = image_to_base64(icon)
        
        else:
            raise TypeError(
                f'`icon` can be `None`, `{Emoji.__name__}`, `bytes-like`, got '
                f'{icon.__class__.__name__}; {reprlib.repr(icon)}.'
            )
        
        
        data = await self.http.role_create(guild_id, data, reason)
        
        return Role(data, guild_id)
    
    
    async def role_move(self, role, position, *, reason=None):
        """
        Moves the given role.
        
        This method is a coroutine.
        
        Parameters
        ----------
        role : ``Role``, `tuple` of (`int`, `int`)
            The role to move.
        position : `int`
            The position to move the given role.
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Shows up at the respective guild's audit logs.
        
        Raises
        ------
        TypeError
            If `role` was not given neither as ``Role`` nor as `tuple` of (`int`, `int`).
        ValueError
            - If default role would be moved.
            - If any role would be moved to position `0`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        if isinstance(role, Role):
            guild_id = role.guild_id
            role_id = 0
        
        else:
            snowflake_pair = maybe_snowflake_pair(role)
            if snowflake_pair is None:
                raise TypeError(
                    f'`role` can be `{Role.__name__}`, `tuple` (`int`, `int`), got '
                    f'{role.__class__.__name__}; {role!r}.'
                )
            
            guild_id, role_id = snowflake_pair
            role = None
        
        guild = GUILDS.get(guild_id, None)
        if (guild is None) or guild.partial:
            guild = await self.guild_sync(guild_id)
        
        if role is None:
            try:
                role = ROLES[role_id]
            except KeyError:
                # Noice
                return
        
        # Is there nothing to move?
        if role.position == position:
            return
        
        # Default role cannot be moved to position not 0
        if role.position == 0:
            if position != 0:
                raise ValueError(
                    f'Default role cannot be moved, got {role!r}.'
                )
        # non default role cannot be moved to position 0
        else:
            if position == 0:
                raise ValueError(
                    f'Role cannot be moved to position `0`, got {role!r}.'
                )
        
        data = change_on_switch(guild.role_list, role, position, key=role_move_key)
        if not data:
            return
        
        await self.http.role_move(guild_id, data, reason)
    
    
    async def _role_reorder_roles_element_validator(self, item):
        """
        Validates a role-position pair.
        
        This method is a coroutine.
        
        Parameters
        ----------
        item : `tuple` (``Role`` or (`tuple` (`int, `int`), `int`) items or `Any`
            A `dict`, `list`, `set`, `tuple`, which contains role-position items.
        
        Returns
        -------
        role : ``Role``
            The validated role.
        guild : ``None`, ``Guild``
            The role's guild.
        
        Yields
        ------
        item : `None`, `tuple` (``Role``, ``Guild``, `int`)
        
        Raises
        ------
        TypeError
            If `item` has invalid format.
        """
        if not isinstance(item, tuple):
            raise TypeError(
                f'`roles` item can be `tuple`, got {item.__class__.__name__}; {item!r}.'
            )
        
        item_length = len(item)
        if item_length != 2:
            raise TypeError(
                f'`roles` item length can be `2`, got {item_length!r}; {item!r}.'
            )
        
        role, position = item
        if isinstance(role, Role):
            guild_id = role.guild_id
            role_id = role.id
        
        else:
            snowflake_pair = maybe_snowflake_pair(role)
            if snowflake_pair is None:
                raise TypeError(
                    f'`roles` item[0] can be `{Role.__name__}`, `tuple` (`int`, `int`), got '
                    f'{role.__class__.__name__}; {role!r}.'
                )
            
            guild_id, role_id = snowflake_pair
            role = None
        
        guild = GUILDS.get(guild_id, None)
        if (guild is None) or guild.partial:
            guild = await self.guild_sync(guild_id)
        
        if role is None:
            role = ROLES.get(role_id, None)
        
        if not isinstance(position, int):
            raise TypeError(
                f'`position` can be `int`, got {position.__class__.__name__}; {position!r}.'
            )
        
        return role, guild, position
    
    
    async def _role_reorder_roles_validator(self, roles):
        """
        Validates `roles` parameter of ``.role_reorder``.
        
        This method is an asynchronous generator.
        
        Parameters
        ----------
        `roles`: (`dict` like or `iterable`) of `tuple` (``Role`` or (`tuple` (`int, `int`), `int`) items
            A `dict`, `list`, `set`, `tuple`, which contains role-position items.
        
        Yields
        ------
        item : `None`, `tuple` (``Role``, ``Guild``, `int`)
        
        Raises
        ------
        TypeError
            If `roles`'s format is not any of the expected ones.
        """
        if isinstance(roles, dict):
            for item in roles.items():
                yield await self._role_reorder_roles_element_validator(item)
        elif isinstance(roles, (list, set, tuple)):
            for item in roles:
                yield await self._role_reorder_roles_element_validator(item)
        else:
            raise TypeError(
                f'`roles` can be `dict-like` with (`{Role.__name__}, `int`) items, or as other '
                f'iterable with (`{Role.__name__}, `int`) elements, got {roles!r}.'
            )
    
    
    async def role_reorder(self, roles, *, reason=None):
        """
        Moves more roles at their guild to the specific positions.
        
        Partial roles are ignored and if passed any, every role's position after it is reduced. If there are roles
        passed with different guilds, then `ValueError` will be raised. If there are roles passed with the same
        position, then their positions will be sorted out.
        
        This method is a coroutine.
        
        Parameters
        ----------
        roles : (`dict` like or `iterable`) of `tuple` (``Role`` or (`tuple` (`int, `int`), `int`) items
            A `dict`, `list`, `set`, `tuple`, which contains role-position items.
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Shows up at the respective guild's audit logs.
        
        Raises
        ------
        TypeError
            If `roles`'s format is not any of the expected ones.
        ValueError
            - If default role would be moved.
            - If any role would be moved to position `0`.
            - If roles from more guilds are passed.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        # Lets check `roles` structure
        roles_valid = []
        
        guild = None
        
        # Is `roles` passed as dict-like?
        async for element in self._role_reorder_roles_validator(roles):
            if element is None:
                continue
            
            role, maybe_guild, position = element
            if maybe_guild is None:
                pass
            if guild is None:
                guild = maybe_guild
            else:
                if guild is not maybe_guild:
                    raise ValueError(
                        f'`roles` are from multiple guilds, got {roles!r}; guild_1={guild!r}; '
                        f'guild_2={guild!r}.'
                    )
            
            roles_valid.append((role, position))
        
        # Nothing to move, nice
        if not roles_valid:
            return
        
        # Check default and moving to default position
        index = 0
        limit = len(roles_valid)
        while True:
            if index == limit:
                break
            
            role, position = roles_valid[index]
            # Default role cannot be moved
            if (role is not None) and (role.position == 0):
                if position != 0:
                    raise ValueError(
                        f'Default role cannot be moved, got {role!r}; roles={roles!r}.'
                    )
                
                # default and moving to default, lets delete it
                del roles_valid[index]
                limit -= 1
                continue
                
            else:
                # Role cannot be moved to default position
                if position == 0:
                    raise ValueError(
                        f'Role cannot be moved to position `0`, got {role!r}; roles={roles!r}.'
                    )
            
            index += 1
            continue
        
        if not limit:
            return
        
        # Check dupe roles
        roles = set()
        ln = 0
        
        for role, position in roles_valid:
            if role is None:
                continue
            
            roles.add(role)
            if len(roles) == ln:
                raise ValueError(
                    f'`{Role.__name__}`: {role!r} is duped, got {roles!r}.'
                )
            
            ln += 1
            continue
        
        # Now that we have the roles, lets order them
        roles_valid.sort(key=role_reorder_valid_roles_sort_key)
        
        # Cut out non roles.
        limit = len(roles_valid)
        index = 0
        negate_position = 0
        while (index < limit):
            role, position = roles_valid[index]
            if role is None:
                del roles_valid[-1]
                limit -= 1
                negate_position += 1
            else:
                
                if negate_position:
                    roles_valid[index] = (role, position - negate_position)
                
                index += 1
        
        # Remove dupe indexes
        index = 0
        limit = len(roles_valid)
        last_position = 0
        increase_position = 0
        while (index < limit):
            role, position = roles_valid[index]
            if position == last_position:
                increase_position += 1
            
            if increase_position:
                roles_valid[index] = (role, position + increase_position)
            
            last_position = position
            index += 1
            continue
        
        
        # Lets cut out every other role from the guild's
        roles_leftover = set(guild.roles.values())
        for item in roles_valid:
            role = item[0]
            roles_leftover.remove(role)
        
        roles_leftover = sorted(roles_leftover)
    
        target_order = []
        
        index_valid = 0
        index_leftover = 0
        limit_valid = len(roles_valid)
        limit_leftover = len(roles_leftover)
        position_target = 0
        
        while True:
            if index_valid == limit_valid:
                while True:
                    if index_leftover == limit_leftover:
                        break
                    
                    role = roles_leftover[index_leftover]
                    index_leftover += 1
                    target_order.append(role)
                    continue
                
                break
            
            if index_leftover == limit_leftover:
                while True:
                    if index_valid == limit_valid:
                        break
                    
                    role = roles_valid[index_valid][0]
                    index_valid += 1
                    target_order.append(role)
                    continue
                
                
                break
            
            role, position = roles_valid[index_valid]
            if position == position_target:
                position_target += 1
                index_valid += 1
                target_order.append(role)
                continue
            
            role = roles_leftover[index_leftover]
            position_target = position_target + 1
            index_leftover = index_leftover + 1
            target_order.append(role)
            continue
        
        data = []
        
        for index, role in enumerate(target_order):
            position = role.position
            if index == position:
                continue
            
            data.append(role_move_key(role, index))
            continue
        
        # Nothing to move
        if not data:
            return
        
        await self.http.role_move(guild.id, data, reason)
    
    
    async def guild_sync_roles(self, guild):
        """
        Deprecated and will be removed in 2022 Dec. Please use ``.guild_role_get_all`` instead.
        """
        warnings.warn(
            (
                f'`{self.__class__.__name__}.guild_sync_roles` is deprecated and will be '
                f'removed in 2022 Dec. Please use `.guild_role_get_all` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        
        return await self.guild_role_get_all(guild)
    
    
    async def guild_role_get_all(self, guild):
        """
        Requests the given guild's roles and if there any de-sync between the wrapper and Discord, applies the
        changes.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild, what's roles will be requested.
        
        Returns
        -------
        roles : `list` of ``Role``
        
        Raises
        ------
        TypeError
            If `guild` was not given neither as ``Guild``, nor as `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild, guild_id = get_guild_and_id(guild)
        
        data = await self.http.guild_role_get_all(guild_id)
        if guild is None:
            guild = create_partial_guild_from_id(guild_id)
        
        guild._sync_roles(data)
        
        return [*guild.roles.values()]



__all__ = ()


from .base import ChannelMetadataBase
from .guild_announcements import ChannelMetadataGuildAnnouncements
from .guild_category import ChannelMetadataGuildCategory
from .guild_directory import ChannelMetadataGuildDirectory
from .guild_forum import ChannelMetadataGuildForum
from .guild_stage import ChannelMetadataGuildStage
from .guild_store import ChannelMetadataGuildStore
from .guild_text import ChannelMetadataGuildText
from .guild_thread_announcements import ChannelMetadataGuildThreadAnnouncements
from .guild_thread_private import ChannelMetadataGuildThreadPrivate
from .guild_thread_public import ChannelMetadataGuildThreadPublic
from .guild_voice import ChannelMetadataGuildVoice
from .private import ChannelMetadataPrivate
from .private_group import ChannelMetadataPrivateGroup


CHANNEL_METADATA_TYPES = {
    ChannelMetadataGuildAnnouncements.type: ChannelMetadataGuildAnnouncements,
    ChannelMetadataGuildCategory.type: ChannelMetadataGuildCategory,
    ChannelMetadataGuildDirectory.type: ChannelMetadataGuildDirectory,
    ChannelMetadataGuildForum.type: ChannelMetadataGuildForum,
    ChannelMetadataGuildStage.type: ChannelMetadataGuildStage,
    ChannelMetadataGuildStore.type: ChannelMetadataGuildStore,
    ChannelMetadataGuildText.type: ChannelMetadataGuildText,
    ChannelMetadataGuildThreadAnnouncements.type: ChannelMetadataGuildThreadAnnouncements,
    ChannelMetadataGuildThreadPrivate.type: ChannelMetadataGuildThreadPrivate,
    ChannelMetadataGuildThreadPublic.type: ChannelMetadataGuildThreadPublic,
    ChannelMetadataGuildVoice.type: ChannelMetadataGuildVoice,
    ChannelMetadataPrivate.type: ChannelMetadataPrivate,
    ChannelMetadataPrivateGroup.type: ChannelMetadataPrivateGroup,
}


def get_channel_metadata_type(channel_type_value):
    """
    Gets metadata type for the given channel type value.
    
    Parameters
    ----------
    channel_type_value : `int`
        The channel's type.
    
    Returns
    -------
    channel_metadata_type : ``ChannelMetadataBase``
    """
    return CHANNEL_METADATA_TYPES.get(channel_type_value, ChannelMetadataBase)



__all__ = ('FormSubmitInteraction', 'FormSubmitInteractionOption',)

import reprlib

from scarletio import copy_docs

from ..components import ComponentType

from .interaction_field_base import InteractionFieldBase


class FormSubmitInteraction(InteractionFieldBase):
    """
    Represents a response to a ``InteractionForm``.
    
    Attributes
    ----------
    custom_id : `None`, `str`
        The forms's custom identifier.
    options : `None`, `tuple` of ``FormSubmitInteractionOption``
        Submitted component values.
    """
    __slots__ = ('custom_id', 'options', )
    
    @copy_docs(InteractionFieldBase.__new__)
    def __new__(cls, data, interaction_event):
        # custom_id
        custom_id = data.get('custom_id', None)
        if (custom_id is not None) and (not custom_id):
            custom_id = None
        
        # options
        option_datas = data.get('components', None)
        if (option_datas is None) or (not option_datas):
            options = None
        else:
            options = tuple(FormSubmitInteractionOption(option_data) for option_data in option_datas)
        
        
        self = object.__new__(cls)
        
        self.custom_id = custom_id
        self.options = options
        
        return self


    @copy_docs(InteractionFieldBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__,]
        
        repr_parts.append(' custom_id=')
        repr_parts.append(reprlib.repr(self.custom_id))
        
        options = self.options
        if (options is not None):
            repr_parts.append(', options=[')
            
            index = 0
            limit = len(options)
            
            while True:
                option = options[index]
                index += 1
                repr_parts.append(repr(option))
                
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append(']')
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(InteractionFieldBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # custom_id
        custom_id = self.custom_id
        if (custom_id is not None):
            hash_value ^= hash(custom_id)
        
        # options
        options = self.options
        if (options is not None):
            hash_value ^= (len(options) << 8)
            
            for option in options:
                hash_value ^= hash(option)
            
        return hash_value
    
    
    @copy_docs(InteractionFieldBase.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        # custom_id
        if self.custom_id != other.custom_id:
            return False
        
        # options
        if self.options != other.options:
            return False
        
        return True
    
    
    def iter_custom_ids_and_values(self):
        """
        Iterates over all the `custom_id`-s and values of the form submit interaction.
        
        This method is an iterable generator.
        
        Yields
        ------
        custom_id : `str`
            The `custom_id` of a represented component.
        value : `str`
            The `value` passed by the user.
        """
        options = self.options
        if (options is not None):
            for option in options:
                yield from option.iter_custom_ids_and_values()
    
    
    def get_custom_id_value_relation(self):
        """
        Returns a dictionary with `custom_id` to `value` relation.
        
        Returns
        -------
        custom_id_value_relation : `dict` of (`str`, `str`) items
        """
        custom_id_value_relation = {}
        
        for custom_id, value in self.iter_custom_ids_and_values():
            if (value is not None):
                custom_id_value_relation[custom_id] = value
        
        return custom_id_value_relation
    
    
    def get_value_for(self, custom_id_to_match):
        """
        Returns the value for the given `custom_id`.
        
        Parameters
        ----------
        custom_id_to_match : `str`
            A respective components `custom_id` to match.
        
        Returns
        -------
        value : `None`, `str`
            The value if any.
        """
        for custom_id, value in self.iter_custom_ids_and_values():
            if (custom_id == custom_id_to_match):
                return value
    
    
    def get_match_and_value(self, matcher):
        """
        Gets a `custom_id`'s value matching the given `matcher`.
        
        Parameters
        ----------
        matcher : `callable`
            Matcher to call on a `custom_id`
            
            Should accept the following parameters:
            
            +-----------+-----------+
            | Name      | Type      |
            +===========+===========+
            | custom_id | `str`     |
            +-----------+-----------+
            
            Should return non-`None` on success.
        
        Returns
        -------
        match : `None`, `Any`
            The returned value by the ``matcher``
        value : `None`, `str`
            The matched `custom_id`'s value.
        """
        for custom_id, value in self.iter_custom_ids_and_values():
            match = matcher(custom_id)
            if (match is not None):
                return match, value
        
        return None, None
    
    
    def iter_matches_and_values(self, matcher):
        """
        Gets a `custom_id`'s value matching the given `matcher`.
        
        This method is an iterable generator.
        
        Parameters
        ----------
        matcher : `callable`
            Matcher to call on a `custom_id`
            
            Should accept the following parameters:
            
            +-----------+-----------+
            | Name      | Type      |
            +===========+===========+
            | custom_id | `str`     |
            +-----------+-----------+
            
            Should return non-`None` on success.
        
        Yields
        -------
        match : `None`, `Any`
            The returned value by the ``matcher``
        value : `None`, `str`
            The matched `custom_id`'s value.
        """
        for custom_id, value in self.iter_custom_ids_and_values():
            match = matcher(custom_id)
            if (match is not None):
                yield match, value


class FormSubmitInteractionOption:
    """
    Attributes
    ----------
    custom_id : `None`, `str`
        The option's respective component's type.
        
    options : `None`, `tuple` of ``FormSubmitInteractionOption``
        Mutually exclusive with the `value` field.
    
    type : ``ComponentType``
        The option respective component's type.
        
    value : `None`, `str`
        Mutually exclusive with the `options` field.
    """
    __slots__ = ('custom_id', 'options', 'type', 'value')
    
    def __new__(cls, data):
        """
        Creates a new ``FormSubmitInteractionOption`` from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            The received form submit interaction option data.
        """
        # custom_id
        custom_id = data.get('custom_id', None)
        if (custom_id is not None) and (not custom_id):
            custom_id = None
        
        # options
        option_datas = data.get('components', None)
        if (option_datas is None) or (not option_datas):
            options = None
        else:
            options = tuple(FormSubmitInteractionOption(option_data) for option_data in option_datas)
        
        # type
        type_ = ComponentType.get(data.get('type', 0))
        
        # value
        value = data.get('value', None)
        if (value is not None) and (not value):
            value = None
        
        self = object.__new__(cls)
        
        self.custom_id = custom_id
        self.options = options
        self.type = type_
        self.value = value
        
        return self
        
    def __repr__(self):
        """Returns the application command interaction option's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        # Descriptive fields : type
        
        # type
        type_ = self.type
        if type_ is not ComponentType.none:
            repr_parts.append(' type=')
            repr_parts.append(type_.name)
            repr_parts.append(' (')
            repr_parts.append(repr(type_.value))
            repr_parts.append(')')
            
            field_added = True
        
        else:
            field_added = False
        
        # System fields : custom_id
        
        # custom_id
        
        if field_added:
            repr_parts.append(',')
        repr_parts.append(' custom_id=')
        repr_parts.append(reprlib.repr(self.custom_id))
        
        # Extra descriptive fields : options | value
        # options
        options = self.options
        if (options is not None):
            repr_parts.append(', options=[')
            
            index = 0
            limit = len(options)
            
            while True:
                option = options[index]
                index += 1
                repr_parts.append(repr(option))
                
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append(']')
        
        # value
        value = self.value
        if (value is not None):
            repr_parts.append(', value=')
            repr_parts.append(repr(value))
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)


    def __hash__(self):
        """Returns the form submit interaction's option hash value."""
        hash_value = 0
        
        # custom_id
        custom_id = self.custom_id
        if (custom_id is not None):
            hash_value ^= hash(custom_id)
        
        # options
        options = self.options
        if (options is not None):
            hash_value ^= (len(options) << 8)
            
            for option in options:
                hash_value ^= hash(option)
        
        # type
        hash_value ^= self.type.value
        
        # value
        value = self.value
        if (value is not None):
            hash_value ^= hash(value)
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns whether the two submit interaction options are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # custom_id
        if self.custom_id != other.custom_id:
            return False
        
        # options
        if self.options != other.options:
            return False
        
        # type
        if self.type is not other.type:
            return False
        
        # value
        if self.value != other.value:
            return False
        
        return True
    
    
    def iter_custom_ids_and_values(self):
        """
        Iterates over all the `custom_id`-s and values of the form submit interaction option.
        
        This method is an iterable generator.
        
        Yields
        ------
        custom_id : `str`
            The `custom_id` of a represented component.
        value : `str`
            The `value` passed by the user.
        """
        custom_id = self.custom_id
        if (custom_id is not None):
            yield custom_id, self.value
        
        options = self.options
        if (options is not None):
            for option in options:
                yield from option.iter_custom_ids_and_values()
